import copy
import datetime
from pathlib import PurePath
from typing import Callable, List

import pytest
from helper.lang import one_moment, retry_assert
from helper.torrent import TorrentData, make_torrent_files, make_torrent_single
from helper.webapi import temporary_torrents

from aioqbt import exc
from aioqbt.api import AddFormBuilder
from aioqbt.api.types import (
    Category,
    ContentLayout,
    FilePriority,
    InfoFilter,
    PieceState,
    StopCondition,
    TorrentInfo,
    TorrentProperties,
    TorrentState,
)
from aioqbt.client import APIClient
from aioqbt.version import APIVersion


@pytest.mark.asyncio
async def test_add(client: APIClient):
    up_limit = 111**1
    dl_limit = 333**1
    category = "add-category"
    savepath = "somewhere/there"

    sample = make_torrent_single("add")
    now = datetime.datetime.now()

    await client.torrents.add(
        AddFormBuilder.with_client(client)
        .savepath(savepath)
        .cookie("dummy=empty")
        .rename("add-renamed")
        .root_folder(True)
        .paused(True)
        .skip_checking(False)  # skip checking cause first last piece prio become False
        .up_limit(up_limit)
        .dl_limit(dl_limit)
        .category(category)
        .sequential_download(True)
        .first_last_piece_prio(True)
        .auto_tmm(False)
        .include_file(sample.data, f"{sample.name}.torrent")
        .build()
    )
    await one_moment()

    torrents = await client.torrents.info(hashes=(sample.hash,))
    assert len(torrents) == 1

    # verify torrent is correctly added
    info = torrents[0]
    assert info is not None, torrents
    assert info.name == "add-renamed"
    assert info.hash == sample.hash
    assert info.state in {TorrentState.CHECKING_RESUME_DATA, TorrentState.PAUSED_DL}
    assert info.up_limit == up_limit
    assert info.dl_limit == dl_limit
    assert info.seq_dl
    assert info.f_l_piece_prio
    assert not info.auto_tmm
    assert info.category == category

    save_path_parts_expected = list(PurePath(savepath).parts)
    save_path_parts_actual = list(PurePath(info.save_path).parts)

    assert save_path_parts_expected == save_path_parts_actual[-len(save_path_parts_expected) :]
    assert abs((info.added_on - now).total_seconds()) <= 10, (info.added_on, now)

    await client.torrents.delete((sample.hash,), True)


@pytest.mark.asyncio
async def test_add_share_limits(client: APIClient):
    if APIVersion.compare(client.api_version, (2, 8, 1)) < 0:
        pytest.skip("API v2.8.1")

    sample = make_torrent_single("add_share_limits")

    ratio_limit = 2
    seeding_time_limit = datetime.timedelta(days=2)

    await client.torrents.add(
        AddFormBuilder.with_client(client)
        .ratio_limit(ratio_limit)
        .seeding_time_limit(seeding_time_limit)
        .include_file(sample.data, f"{sample.name}.torrent")
        .build()
    )

    @retry_assert
    async def assert_share_limits():
        torrents = await client.torrents.info(hashes=(sample.hash,))
        assert len(torrents) == 1
        info = torrents[0]

        assert info.seeding_time_limit == seeding_time_limit
        assert info.ratio_limit == ratio_limit

    await assert_share_limits()
    await client.torrents.delete((sample.hash,), True)


@pytest.mark.asyncio
async def test_add_stop_condition(client: APIClient):
    if APIVersion.compare(client.api_version, (2, 8, 15)) < 0:
        pytest.skip("API v2.8.15")

    sample = make_torrent_single("add_stop_condition_sample")

    await client.torrents.add(
        AddFormBuilder.with_client(client)
        .stop_condition(StopCondition.METADATA_RECEIVED)
        .include_url(sample.magnet)
        .build()
    )

    @retry_assert
    async def assert_no_metadata():
        torrents = await client.torrents.info(hashes=(sample.hash,))
        assert len(torrents) == 1
        info = torrents[0]
        assert info.state == TorrentState.META_DL, info.state

    await assert_no_metadata()

    try:
        await client.torrents.add(
            AddFormBuilder.with_client(client)
            .stop_condition(StopCondition.METADATA_RECEIVED)
            .include_file(sample.data, f"{sample.name}.torrent")
            .build()
        )
    except exc.AddTorrentError:
        # As of v4.5.2, "torrents/add" fails if torrent is already added
        # However, metadata is merged if available.
        pass

    @retry_assert
    async def assert_state():
        torrents = await client.torrents.info(hashes=(sample.hash,))
        assert len(torrents) == 1
        info = torrents[0]
        assert info.state in {TorrentState.CHECKING_RESUME_DATA, TorrentState.PAUSED_DL}

    await assert_state()


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "name,layout,make_torrent,parts_count",
    (
        ("subfolder_single", ContentLayout.SUBFOLDER, make_torrent_single, 2),
        ("nosubfolder_multiple", ContentLayout.NO_SUBFOLDER, make_torrent_files, 2),
    ),
)
async def test_add_content_layout(
    client: APIClient,
    name: str,
    layout: ContentLayout,
    make_torrent: Callable[[str], TorrentData],
    parts_count: int,
):
    if APIVersion.compare(client.api_version, (2, 7, 0)) < 0:
        pytest.skip("API v2.7.0")

    sample = make_torrent(f"add_content_layout_{name!s}")

    await client.torrents.add(
        AddFormBuilder.with_client(client)
        .content_layout(layout)
        .include_file(sample.data, f"{sample.name}.torrent")
        .build()
    )

    @retry_assert(exc_types=exc.NotFoundError)
    async def get_files():
        return await client.torrents.files(hash=sample.hash)

    files = await get_files()
    path = PurePath(files[0].name)
    assert len(path.parts) == parts_count, path
    await client.torrents.delete((sample.hash,), delete_files=True)


@pytest.mark.asyncio
async def test_add_download_path(client: APIClient):
    if APIVersion.compare(client.api_version, (2, 8, 4)) < 0:
        pytest.skip("API v2.8.4")

    sample = make_torrent_single("add_download_path")
    pathname = "apple"

    await client.torrents.add(
        AddFormBuilder.with_client(client)
        .download_path(pathname)
        .use_download_path(True)
        .include_file(sample.data, f"{sample.name}.torrent")
        .build()
    )

    @retry_assert
    async def assert_download_path():
        torrents = await client.torrents.info(hashes=(sample.hash,))
        assert len(torrents) == 1
        info = torrents[0]

        assert hasattr(info, "download_path")
        assert PurePath(info.download_path).name == pathname, info.download_path

    await assert_download_path()


@pytest.mark.asyncio
async def test_add_mixed(client: APIClient):
    """Add torrents with mixed of data and urls"""
    sample_data1 = make_torrent_single("add_mixed_data1")
    sample_data2 = make_torrent_single("add_mixed_data2")
    sample_magnet1 = make_torrent_single("add_mixed_magnet1")
    sample_magnet2 = make_torrent_single("add_mixed_magnet2")
    sample_hash1 = make_torrent_single("add_mixed_hash1")
    sample_hash2 = make_torrent_single("add_mixed_hash2")

    await client.torrents.add(
        AddFormBuilder.with_client(client)
        .paused(True)
        .include_file(sample_data1.data, "data1.torrent")
        .include_url(sample_magnet1.magnet)
        .include_url(sample_hash1.hash)
        .include_file(sample_data2.data, "data2.torrent")
        .include_url(sample_magnet2.magnet)
        .include_url(sample_hash2.hash)
        .build()
    )
    await one_moment()

    hashes = (
        sample_data1.hash,
        sample_data2.hash,
        sample_magnet1.hash,
        sample_magnet2.hash,
        sample_hash1.hash,
        sample_hash2.hash,
    )
    torrents = await client.torrents.info(hashes=hashes)
    assert len(torrents) == len(hashes)

    await client.torrents.delete(hashes, True)


def test_add_builder():
    builder = AddFormBuilder()
    builder_copy = copy.copy(builder)

    assert builder == builder_copy

    builder2 = builder.include_url("0" * 40)

    assert builder != builder2


@pytest.mark.asyncio
async def test_add_error(client: APIClient):
    with pytest.raises(exc.AddTorrentError):
        await client.torrents.add(AddFormBuilder.with_client(client).build())


@pytest.mark.asyncio
async def test_delete_error(client: APIClient):
    with pytest.raises(ValueError):
        await client.torrents.delete("0" * 40, True)


@pytest.mark.asyncio
async def test_count(client: APIClient):
    if APIVersion.compare(client.api_version, (2, 9, 3)) < 0:
        pytest.skip("API 2.9.3")
    count = await client.torrents.count()
    assert isinstance(count, int)


@pytest.mark.asyncio
async def test_torrent_metadata(client: APIClient):
    sample = make_torrent_single("torrent_metadata")
    info_hash = sample.hash
    now = datetime.datetime.now()

    async with temporary_torrents(client, sample):
        # torrents.info()
        torrents = await client.torrents.info(hashes=(info_hash,))
        assert isinstance(torrents, list)
        assert len(torrents) == 1
        assert isinstance(torrents[0], TorrentInfo)

        @retry_assert
        async def assert_torrent(info: TorrentInfo):
            # some methods may be unavailable when torrent created.
            # retry asserts if needed.

            assert isinstance(info, TorrentInfo)
            assert isinstance(repr(info), str)
            assert abs((info.added_on - now).total_seconds()) <= 10

            # torrents.properties()
            props = await client.torrents.properties(info_hash)
            assert isinstance(props, TorrentProperties)
            assert props.hash, f"hash={getattr(props, 'hash', None)!r}"
            assert isinstance(repr(props), str)
            assert abs((props.addition_date - now).total_seconds()) <= 10

            # torrents.webseeds()
            webseeds = await client.torrents.webseeds(info_hash)
            assert isinstance(webseeds, list)
            assert len(webseeds) == 0

            # torrents.piece_states
            piece_states = await client.torrents.piece_states(info_hash)
            assert isinstance(piece_states, list)
            assert len(piece_states)
            assert all(s == PieceState.UNAVAILABLE for s in piece_states)

            # torrents.piece_hashes
            piece_hashes = await client.torrents.piece_hashes(info_hash)
            assert isinstance(piece_hashes, list)
            assert len(piece_hashes)
            assert len(piece_hashes) == len(piece_states)
            assert all(isinstance(s, str) and len(s) == 40 for s in piece_hashes)

        await assert_torrent(torrents[0])


@pytest.mark.asyncio
async def test_manipulation(client: APIClient):
    sample = make_torrent_single("manipulation")
    new_name = "manipulation2"
    info_hash = sample.hash
    hashes = (info_hash,)

    async with temporary_torrents(client, sample):
        await client.torrents.rename(sample.hash, new_name)

        async def torrent_info() -> TorrentInfo:
            return (await client.torrents.info(hashes=hashes))[0]

        @retry_assert
        async def assert_name():
            info = await torrent_info()
            assert info.name == new_name

        await assert_name()

        # torrents.resume
        @retry_assert
        async def assert_resumed():
            assert len(await client.torrents.info(filter=InfoFilter.RESUMED, hashes=hashes))

        await client.torrents.resume(hashes)
        await assert_resumed()

        @retry_assert
        async def assert_forced_start():
            assert (await torrent_info()).force_start

        await client.torrents.set_force_start(hashes, True)
        await assert_forced_start()

        @retry_assert
        async def assert_super_seeding():
            assert (await torrent_info()).super_seeding

        await client.torrents.set_super_seeding(hashes, True)
        await assert_super_seeding()

        # torrents.reannounce
        await client.torrents.reannounce(hashes)
        await one_moment()

        # torrents.add_peers
        assert client.api_version is not None
        if client.api_version >= (2, 3, 0):
            await client.torrents.add_peers(hashes, ("127.0.0.2:8080", "127.0.0.3:8080"))
            await one_moment()

        # torrents.pause
        @retry_assert
        async def assert_paused():
            assert len(await client.torrents.info(filter=InfoFilter.PAUSED, hashes=hashes))

        await client.torrents.pause(hashes=hashes)
        await assert_paused()

        # torrents.set_location
        info = await torrent_info()
        addition = "magic-directory"
        new_save_path = f"{info.save_path}/{addition}"
        await client.torrents.set_location(hashes, new_save_path)

        @retry_assert
        async def assert_save_path_changed():
            info = await torrent_info()
            save_path = PurePath(info.save_path)
            assert save_path.name == addition

        await assert_save_path_changed()

        # torrents.recheck
        await client.torrents.recheck(hashes)
        await one_moment()


@pytest.mark.asyncio
async def test_save_path(client: APIClient):
    if APIVersion.compare(client.api_version, (2, 8, 4)) < 0:
        pytest.skip("set_save_path/set_download_path need API v2.8.4")

    sample = make_torrent_single("paths")
    info_hash = sample.hash
    hashes = (info_hash,)

    async with temporary_torrents(client, sample):
        # no need to check resultant path. no error is good.

        # torrents.set_location
        save_path = "paths/save_path"
        await client.torrents.set_save_path(hashes, save_path)

        # torrents.set_save_path
        download_path = "paths/download_path"
        await client.torrents.set_download_path(hashes, download_path)


@pytest.mark.asyncio
async def test_rename_file(client: APIClient):
    if APIVersion.compare(client.api_version, (2, 4, 0)) < 0:
        pytest.skip("rename_file need API v2.4.0")

    sample = make_torrent_files("rename_file")

    async with temporary_torrents(client, sample):
        files = await client.torrents.files(sample.hash)
        assert len(files) == 5

        index = 3
        old_path = PurePath(files[index].name)
        new_path = old_path.with_name("renamed.txt")
        assert old_path.as_posix() == "rename_file/files/03.txt"

        if APIVersion.compare(client.api_version, (2, 8, 0)) < 0:
            # rename_file(hash, index, new_name) if API < v2.8.0
            await client.torrents.rename_file(
                sample.hash,
                index,
                new_path.name,
            )
        else:
            # rename_file(hash, old_path, new_path) if API >= v2.8.0
            await client.torrents.rename_file(
                sample.hash,
                old_path.as_posix(),
                new_path.as_posix(),
            )

        @retry_assert
        async def assert_filename():
            files = await client.torrents.files(sample.hash)
            assert files[index].name == new_path.as_posix()

        await assert_filename()


@pytest.mark.asyncio
async def test_rename_folder(client: APIClient):
    if APIVersion.compare(client.api_version, (2, 8, 0)) < 0:
        pytest.skip("require API 2.8.0")

    sample = make_torrent_files("rename_folder")

    async with temporary_torrents(client, sample):
        # paths consist of torrent names and file names
        await client.torrents.rename_folder(
            sample.hash,
            "rename_folder/files",
            "rename_folder/other",
        )

        @retry_assert
        async def assert_renamed():
            files = await client.torrents.files(sample.hash)
            empty = [s.name for s in files if "other" not in s.name]
            assert empty == []

        await assert_renamed()


@pytest.mark.asyncio
async def test_torrent_auto_tmm(client: APIClient):
    sample = make_torrent_single("torrent_auto_tmm")
    hashes = (sample.hash,)

    async with temporary_torrents(client, sample) as torrents:
        info = torrents[0]
        auto_tmm = info.auto_tmm
        await client.torrents.set_auto_management(hashes, not auto_tmm)
        await one_moment()

        info = (await client.torrents.info(hashes=hashes))[0]
        assert auto_tmm != info.auto_tmm


@pytest.mark.asyncio
async def test_torrent_seq_dl(client: APIClient):
    sample = make_torrent_single("torrent_seq_dl")
    hashes = (sample.hash,)

    async with temporary_torrents(client, sample) as torrents:
        info = torrents[0]
        seq_dl = info.seq_dl
        await client.torrents.toggle_sequential_download(hashes)
        await one_moment()

        info = (await client.torrents.info(hashes=hashes))[0]
        assert seq_dl != info.seq_dl

        await client.torrents.set_sequential_download(hashes, seq_dl)
        await one_moment()

        info = (await client.torrents.info(hashes=hashes))[0]
        assert seq_dl == info.seq_dl

        await client.torrents.set_sequential_download(hashes, seq_dl)
        await one_moment()

        info = (await client.torrents.info(hashes=hashes))[0]
        assert seq_dl == info.seq_dl


@pytest.mark.asyncio
async def test_torrent_flp_prio(client: APIClient):
    sample = make_torrent_single("torrent_flp_prio")
    hashes = (sample.hash,)

    async with temporary_torrents(client, sample) as torrents:
        info = torrents[0]
        flp_prio = info.f_l_piece_prio
        await client.torrents.toggle_first_last_piece_prio(hashes)
        await one_moment()

        info = (await client.torrents.info(hashes=hashes))[0]
        assert flp_prio != info.f_l_piece_prio

        await client.torrents.set_first_last_piece_prio(hashes, flp_prio)
        await one_moment()

        info = (await client.torrents.info(hashes=hashes))[0]
        assert flp_prio == info.f_l_piece_prio

        await client.torrents.set_first_last_piece_prio(hashes, flp_prio)
        await one_moment()

        info = (await client.torrents.info(hashes=hashes))[0]
        assert flp_prio == info.f_l_piece_prio


@pytest.mark.asyncio
async def test_limits(client: APIClient):
    dl_limit = 111
    up_limit = 222
    ratio_limit = 2
    seeding_time_limit = datetime.timedelta(days=1)
    inactive_seeding_time_limit = datetime.timedelta(days=30)

    sample = make_torrent_single("limits")
    info_hash = sample.hash

    async with temporary_torrents(client, sample) as [info]:
        has_inactive_seeding_time_limit = hasattr(info, "inactive_seeding_time_limit")

        # set dl_limit, up_limit, share_limits
        await client.torrents.set_download_limit((info_hash,), dl_limit)
        await client.torrents.set_upload_limit((info_hash,), up_limit)

        await client.torrents.set_share_limits(
            (info_hash,),
            ratio_limit=ratio_limit,
            seeding_time_limit=seeding_time_limit,
            inactive_seeding_time_limit=inactive_seeding_time_limit,
        )

        await one_moment()

        # check limits
        info = (await client.torrents.info(hashes=(info_hash,)))[0]
        assert info.dl_limit == dl_limit
        assert info.up_limit == up_limit
        assert info.ratio_limit == ratio_limit
        assert info.seeding_time_limit == seeding_time_limit

        if has_inactive_seeding_time_limit:
            assert info.inactive_seeding_time_limit == inactive_seeding_time_limit

        # torrents.download_limit()
        download_limits = await client.torrents.download_limit((info_hash,))
        assert info_hash in download_limits
        assert download_limits[info_hash] == dl_limit

        # torrents.upload_limit()
        upload_limits = await client.torrents.upload_limit((info_hash,))
        assert info_hash in upload_limits
        assert upload_limits[info_hash] == up_limit


@pytest.mark.asyncio
async def test_trackers(client: APIClient):
    tracker = f"{client.base_url}/missing/tracker"
    tracker2 = f"{client.base_url}/missing/tracker2"
    tracker3 = f"{client.base_url}/missing/tracker3"

    sample = make_torrent_single("trackers")
    info_hash = sample.hash

    async with temporary_torrents(client, sample):
        # torrents.trackers
        trackers = await client.torrents.trackers(info_hash)
        assert isinstance(trackers, list)
        assert len(trackers) == 3
        assert all(s.is_special() for s in trackers)

        # torrents.add_trackers
        await client.torrents.add_trackers(info_hash, (tracker, tracker2))
        await one_moment()

        await client.torrents.edit_tracker(info_hash, tracker2, tracker3)
        await one_moment()

        await client.torrents.remove_trackers(info_hash, (tracker,))
        await one_moment()

        trackers = await client.torrents.trackers(info_hash)
        assert len(trackers) == 4
        assert any(s.url == tracker3 for s in trackers)


@pytest.mark.asyncio
async def test_queueing(client: APIClient):
    sample1 = make_torrent_single("queueing1")
    sample2 = make_torrent_single("queueing2")
    sample3 = make_torrent_single("queueing3")
    hashes = (sample1.hash, sample2.hash, sample3.hash)

    # make sure queueing is on or the following would fail
    await client.app.set_preferences(
        {
            "queueing_enabled": True,
        }
    )

    async with temporary_torrents(client, sample1, sample2, sample3):
        h1, h2, h3 = hashes

        async def queued_hashes() -> List[str]:
            torrents = await client.torrents.info(hashes=hashes)
            torrents.sort(key=lambda s: s.priority)
            return [s.hash for s in torrents]

        @retry_assert
        async def assert_order123():
            assert [h1, h2, h3] == await queued_hashes()

        @retry_assert
        async def assert_order312():
            assert [h3, h1, h2] == await queued_hashes()

        await client.torrents.top_prio((h3,))
        await client.torrents.top_prio((h2,))
        await client.torrents.top_prio((h1,))
        # order: 1, 2, 3

        await assert_order123()

        await client.torrents.increase_prio((h2,))
        # order: 2, 1, 3

        await client.torrents.decrease_prio((h1,))
        # order: 2, 3, 1

        await client.torrents.bottom_prio((h2,))
        # order: 3, 1, 2

        await assert_order312()


@pytest.mark.asyncio
async def test_file_priorities(client: APIClient):
    sample = make_torrent_files("file_priorities")
    info_hash = sample.hash

    async def file_priorities() -> List[int]:
        return [s.priority for s in await client.torrents.files(info_hash)]

    async with temporary_torrents(client, sample):
        assert len(await file_priorities()) == 5

        @retry_assert
        async def assert_all_normal():
            expected = [FilePriority.NORMAL] * 5
            assert expected == await file_priorities()

        @retry_assert
        async def assert_file_properties():
            expected = [
                FilePriority.NORMAL,
                FilePriority.NORMAL,
                FilePriority.NO_DOWNLOAD,
                FilePriority.NORMAL,
                FilePriority.NORMAL,
            ]
            assert expected == await file_priorities()

        # libtorrent default priority is 4 which is invalid in qbittorrent
        # so set all to NORMAL first
        await client.torrents.file_prio(info_hash, range(5), FilePriority.NORMAL)
        await assert_all_normal()

        # modify some of them
        await client.torrents.file_prio(info_hash, [2], FilePriority.NO_DOWNLOAD)
        await assert_file_properties()


@pytest.mark.asyncio
async def test_categories(client: APIClient):
    sample = make_torrent_single("categories")
    info_hash = sample.hash
    category_name = "category_name"
    category_path = "category_path"

    # Create category
    await client.torrents.create_category(category_name, category_path)
    await one_moment()

    # Get categories
    categories = await client.torrents.categories()
    assert isinstance(categories, dict)
    assert category_name in categories
    assert isinstance(categories[category_name], Category)
    assert categories[category_name].savePath == category_path

    # Edit category save path
    await client.torrents.edit_category(category_name, "")
    await one_moment()

    categories = await client.torrents.categories()
    assert categories[category_name].savePath == ""

    async with temporary_torrents(client, sample):
        await client.torrents.set_category((info_hash,), category_name)
        await one_moment()

        info = (await client.torrents.info(hashes=(info_hash,)))[0]
        assert info.category == category_name

    # Remove category
    await client.torrents.remove_categories((category_name,))
    await one_moment()

    categories = await client.torrents.categories()
    assert category_name not in categories


@pytest.mark.asyncio
async def test_tags(client: APIClient):
    assert client.api_version is not None
    if client.api_version < (2, 3, 0):
        pytest.skip("tags methods are available since API v2.3.0")

    sample = make_torrent_single("tags")
    info_hash = sample.hash

    async with temporary_torrents(client, sample):
        add_tags = ["dragon-fruits", "figs", "grapes"]
        create_tags = ["apples", "bananas", "cherries"]

        # add tags
        await client.torrents.add_tags((info_hash,), add_tags)
        await one_moment()

        # remove tags
        await client.torrents.remove_tags((info_hash,), ("figs",))
        await one_moment()

        # create tags
        await client.torrents.create_tags(create_tags)
        await one_moment()

        # delete tags
        await client.torrents.delete_tags(("bananas", "dragon-fruits"))
        await one_moment()

        # tags
        expected_tags = {"figs", "grapes", "apples", "cherries"}
        tags = await client.torrents.tags()
        assert isinstance(tags, list)
        assert set(tags) == expected_tags

        await client.torrents.delete_tags(sorted(expected_tags))


@pytest.mark.asyncio
async def test_export(client: APIClient):
    if APIVersion.compare(client.api_version, (2, 8, 11)) < 0:
        pytest.skip("export need API v2.8.11")

    sample = make_torrent_single("export")
    info_hash = sample.hash

    async with temporary_torrents(client, sample):
        data = await client.torrents.export(info_hash)
        assert isinstance(data, bytes)
        assert b"4:infod" in data



import json
from typing import List, Mapping, Optional

from aioqbt._paramdict import ParamDict
from aioqbt.api.types import BuildInfo, NetworkInterface, Preferences
from aioqbt.client import APIGroup

__all__ = ("AppAPI",)


class AppAPI(APIGroup):
    """
    API methods under ``app``.
    """

    async def version(self) -> str:
        """qBittorrent version."""
        return await self._request_text(
            "GET",
            "app/version",
        )

    async def webapi_version(self) -> str:
        """WebUI API version."""
        return await self._request_text(
            "GET",
            "app/webapiVersion",
        )

    async def build_info(self) -> BuildInfo:
        """Build information."""
        # API 2.3.0

        return await self._request_mapped_object(
            BuildInfo,
            "GET",
            "app/buildInfo",
        )

    async def shutdown(self) -> None:
        """Shut down qBittorrent client."""
        await self._request_text(  # pragma: no cover
            "POST",
            "app/shutdown",
        )

    async def preferences(self) -> Preferences:
        """Get application preferences."""
        return await self._request_json(  # type: ignore[no-any-return]
            "GET",
            "app/preferences",
        )

    async def set_preferences(self, prefs: Mapping[str, object]) -> None:
        """
        Set application preferences.

        :param prefs: a mapping of preferences to update.
        """
        prefs = dict(prefs)
        # plus sign (+) were not decoded as space in v4.1.5 or earlier.
        # JSON.dumps() are invalid with default separators argument.
        # Removing spaces in separators allows parsing JSON correctly
        # though plus signs are still not decoded as spaces.
        # This should behave similarly to WebUI.
        # https://github.com/qbittorrent/qBittorrent/issues/10451
        data = {
            "json": json.dumps(prefs, separators=(",", ":")),
        }

        await self._request_text(
            "POST",
            "app/setPreferences",
            data=data,
        )

    async def default_save_path(self) -> str:
        """Default save path for storing downloaded files"""
        return await self._request_text(
            "GET",
            "app/defaultSavePath",
        )

    async def network_interface_list(self) -> List[NetworkInterface]:
        """Network interfaces."""
        # since v4.2.0, API v2.3.0
        return await self._request_mapped_list(
            NetworkInterface,
            "GET",
            "app/networkInterfaceList",
        )

    async def network_interface_address_list(self, iface: Optional[str] = None) -> List[str]:
        """Network addresses."""
        # since v4.2.0, API v2.3.0
        params = ParamDict()
        params.put("iface", iface, optional=False, prepare=str, default="")

        return await self._request_json(  # type: ignore[no-any-return]
            "GET",
            "app/networkInterfaceAddressList",
            params=params,
        )



import time
from typing import TYPE_CHECKING, Any, Dict

if TYPE_CHECKING:

    def prettyPrinter(dictionary: Dict[str, Any]) -> None:
        pass

else:
    from novaprinter import prettyPrinter


def _image_dict(name: str, info_hash: str, size: str) -> Dict[str, str]:
    return {
        "link": f"magnet:?xt=urn:btih:{info_hash}",
        "name": name,
        "size": size,
        "seeds": "-1",
        "leech": "-1",
        "engine_url": "-1",
    }


_TABLE = [
    _image_dict(
        "ubuntu-22.04.3-desktop-amd64.iso",
        "75439d5de343999ab377c617c2c647902956e282",
        "4.69 GiB",
    ),
    _image_dict(
        "ubuntu-20.04.6-desktop-amd64.iso",
        "5f5e8848426129ab63cb4db717bb54193c1c1ad7",
        "4.05 GiB",
    ),
    _image_dict(
        "debian-12.1.0-amd64-netinst.iso",
        "a9164e99d5181cfef0c23c209334103619080908",
        "627.0 MiB",
    ),
]


class linuxsearch:
    url = "http://example.com"
    name = "Linux Search"
    supported_categories = {
        "all": "0",
        "software": "3",
    }

    def search(self, what: str, cat: str = "all") -> None:
        for item in _TABLE:
            name = item["name"]
            if what in name:
                prettyPrinter(item)

        # keep search status in "Running" from "Stopped"
        time.sleep(60)



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import os
import sys

_HERE = os.path.dirname(__file__)
_PROJECT_ROOT = os.path.abspath(os.path.join(_HERE, "../.."))

sys.path.append(os.path.join(_HERE, "_ext"))

import aioqbt

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "aioqbt"
copyright = "2022, Aaron"
author = "Aaron"
version = aioqbt.__version__
release = version

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.duration",
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.extlinks",
    "sphinx.ext.viewcode",
    "_sphinx_monkeypatch",
]

templates_path = ["_templates"]
exclude_patterns = []

# -- Options for intersphinx
intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "aiohttp": ("https://docs.aiohttp.org/en/stable/", None),
}

# -- Options for Python ------------------------------------------------------
python_display_short_literal_types = False

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]
html_theme_options = {
    "style_external_links": False,
    "navigation_depth": 3,
}
html_context = {
    "display_github": True,
    "github_user": "tsangwpx",
    "github_repo": "aioqbt",
    "github_version": "main",
    "conf_py_path": "/docs/source/",
}
html_css_files = [
    "custom.css",
]

# -- Options for extlinks
extlinks = {
    "issue": ("https://github.com/tsangwpx/aioqbt/issues/%s", "issue %s"),
    "APIWiki": (
        "https://github.com/qbittorrent/qBittorrent/wiki/WebUI-API-(qBittorrent-4.1)%s",
        "[APIWiki %s]",
    ),
}

# -- Options for autodoc
autodoc_member_order = "bysource"



import asyncio
import functools
import logging
import random
import sys
import time
from typing import Any, Awaitable, Callable, Optional, Tuple, Type, TypeVar, Union, overload

from typing_extensions import ParamSpec


async def one_moment(duration: float = 1 / 1000) -> None:
    """Pause for a moment"""
    await asyncio.sleep(duration)


async def busy_wait_for(
    predicate: Callable[[], Awaitable[bool]],
    timeout: float = 10,
    step: Optional[float] = None,
) -> bool:
    if step is None:
        step = max(timeout / 100, 0.01)

    now = time.monotonic()
    deadline = now + timeout

    while now < deadline:
        try:
            result = await asyncio.wait_for(predicate(), deadline - now)
        except asyncio.TimeoutError:
            break

        if result:
            return True

        await asyncio.sleep(step)
        now = time.monotonic()

    return False


if sys.version_info >= (3, 9):

    def randbytes(r: random.Random, n: int) -> bytes:
        return r.randbytes(n)

else:

    def randbytes(r: random.Random, n: int) -> bytes:
        return r.getrandbits(n * 8).to_bytes(n, "little")


Fn = TypeVar("Fn", bound=Callable[..., Any])
T = TypeVar("T")
P = ParamSpec("P")


@overload
def retry_assert(
    *,
    max_attempts: int = ...,
    pause: Optional[float] = ...,
    delay: Optional[float] = ...,
    exc_types: Union[Type[BaseException], Tuple[Type[BaseException], ...], None] = ...,
    logger: Optional[logging.Logger] = ...,
) -> Callable[[Fn], Fn]:
    pass


@overload
def retry_assert(
    fn: Fn,
    *,
    max_attempts: int = ...,
    pause: Optional[float] = ...,
    delay: Optional[float] = ...,
    exc_types: Union[Type[BaseException], Tuple[Type[BaseException], ...], None] = ...,
    logger: Optional[logging.Logger] = ...,
) -> Fn:
    pass


def retry_assert(
    fn: Optional[Callable[P, Awaitable[T]]] = None,
    *,
    max_attempts: int = 4,
    pause: Optional[float] = None,
    delay: Optional[float] = None,
    exc_types: Union[Type[BaseException], Tuple[Type[BaseException], ...], None] = None,
    logger: Optional[logging.Logger] = None,
) -> Union[Callable[[Fn], Fn], Callable[P, Awaitable[T]]]:
    """
    Retry a coroutine if AssertionError raise.

    :param fn: function being decorated
    :param int max_attempts: maximum number of attempts
    :param float|None delay: initial delay
    :param float|None pause: pause between attempts
    :param exc_types: exception types to catch. The default value is `AssertionError`
    :param logger: logger to record failed attempt
    """

    if max_attempts <= 0:
        raise ValueError("max_attempts <= 0")

    if fn is None:
        return functools.partial(  # type: ignore[return-value]
            retry_assert,
            pause=pause,
            delay=delay,
            max_attempts=max_attempts,
            exc_types=exc_types,
            logger=logger,
        )

    if pause is None:
        pause = 1

    if delay is None:
        delay = 0

    if exc_types is None:
        exc_types = AssertionError

    @functools.wraps(fn)
    async def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
        assert pause is not None
        assert delay is not None

        await asyncio.sleep(delay)

        last_exception: Optional[BaseException] = None

        for count in range(1, max_attempts + 1):
            try:
                return await fn(*args, **kwargs)
            except exc_types as ex:
                last_exception = ex

                if logger is not None:
                    logger.warning("Attempt %d/%d failed", count, max_attempts, exc_info=True)

            if count < max_attempts:
                await asyncio.sleep(pause)

        assert last_exception is not None

        raise last_exception

    return wrapper



import pytest

from aioqbt.api.types import LogMessage, LogPeer
from aioqbt.client import APIClient


@pytest.mark.asyncio
async def test_log(client: APIClient):
    logs = await client.log.main()
    assert isinstance(logs, list)
    if logs:
        message = logs[0]
        assert isinstance(message, LogMessage)
        assert isinstance(message.id, int)

    peers = await client.log.peers()
    assert isinstance(peers, list)
    if peers:
        peer_log = peers[0]
        assert isinstance(peer_log, LogPeer)
        assert isinstance(peer_log.id, LogPeer)



import pytest
from helper.lang import retry_assert

from aioqbt.api.types import BuildInfo, NetworkInterface, Preferences
from aioqbt.client import APIClient
from aioqbt.version import APIVersion


@pytest.mark.asyncio
async def test_app(client: APIClient):
    version = await client.app.version()
    assert isinstance(version, str)
    assert version.startswith("v4.")

    webapi_version = await client.app.webapi_version()
    assert isinstance(webapi_version, str)
    assert webapi_version.startswith("2.")

    if APIVersion.compare(client.api_version, (2, 3, 0)) >= 0:
        build_info = await client.app.build_info()
        assert isinstance(build_info, BuildInfo)
        assert isinstance(build_info.qt, str)
        assert isinstance(build_info.libtorrent, str)
        assert isinstance(build_info.openssl, str)

    default_save_path = await client.app.default_save_path()
    assert isinstance(default_save_path, str)


@pytest.mark.asyncio
async def test_preferences(client: APIClient):
    prefs: Preferences
    prefs = await client.app.preferences()

    assert isinstance(prefs, dict), type(prefs)


@pytest.mark.asyncio
async def test_set_preferences(client: APIClient):
    changes = Preferences()
    changes["alt_dl_limit"] = 7 * 1024

    await client.app.set_preferences(changes)

    @retry_assert
    async def assert_updated():
        prefs = await client.app.preferences()
        assert prefs["alt_dl_limit"] == changes["alt_dl_limit"], prefs["alt_dl_limit"]

    await assert_updated()


@pytest.mark.asyncio
async def test_interfaces(client: APIClient):
    if APIVersion.compare(client.api_version, (2, 3, 0)) < 0:
        pytest.skip("networkInterfaceList are available since API v2.3.0")

    interfaces = await client.app.network_interface_list()
    assert isinstance(interfaces, list)

    for iface in interfaces:
        assert isinstance(iface, NetworkInterface)
        assert isinstance(iface.name, str)
        assert isinstance(iface.value, str)

    addresses = await client.app.network_interface_address_list()
    assert isinstance(addresses, list)

    for addr in addresses:
        assert isinstance(addr, str)



"""
API methods by group.

See also :APIWiki:`the WebUI API reference <>` on qBittorrent wiki.
"""

# re-export API groups, constants, and types here

from aioqbt.api.app import AppAPI as AppAPI
from aioqbt.api.auth import AuthAPI as AuthAPI
from aioqbt.api.log import LogAPI as LogAPI
from aioqbt.api.rss import RSSAPI as RSSAPI
from aioqbt.api.search import SearchAPI as SearchAPI
from aioqbt.api.sync import SyncAPI as SyncAPI
from aioqbt.api.torrents import AddFormBuilder as AddFormBuilder
from aioqbt.api.torrents import TorrentsAPI as TorrentsAPI
from aioqbt.api.transfer import TransferAPI as TransferAPI

# isort: split
from aioqbt.api.types import ConnectionStatus as ConnectionStatus
from aioqbt.api.types import ContentLayout as ContentLayout
from aioqbt.api.types import FilePriority as FilePriority
from aioqbt.api.types import InactiveSeedingTimeLimits as InactiveSeedingTimeLimits
from aioqbt.api.types import InfoFilter as InfoFilter
from aioqbt.api.types import PieceState as PieceState
from aioqbt.api.types import RatioLimits as RatioLimits
from aioqbt.api.types import SeedingTimeLimits as SeedingTimeLimits
from aioqbt.api.types import StopCondition as StopCondition
from aioqbt.api.types import TorrentState as TorrentState
from aioqbt.api.types import TrackerStatus as TrackerStatus

# isort: split
from aioqbt.api.types import BuildInfo as BuildInfo
from aioqbt.api.types import Category as Category
from aioqbt.api.types import FileEntry as FileEntry
from aioqbt.api.types import LogMessage as LogMessage
from aioqbt.api.types import LogPeer as LogPeer
from aioqbt.api.types import NetworkInterface as NetworkInterface
from aioqbt.api.types import Preferences as Preferences
from aioqbt.api.types import RSSArticle as RSSArticle
from aioqbt.api.types import RSSFeed as RSSFeed
from aioqbt.api.types import RSSFolder as RSSFolder
from aioqbt.api.types import RSSItem as RSSItem
from aioqbt.api.types import RSSRule as RSSRule
from aioqbt.api.types import SearchJobResults as SearchJobResults
from aioqbt.api.types import SearchJobStart as SearchJobStart
from aioqbt.api.types import SearchJobStatus as SearchJobStatus
from aioqbt.api.types import SearchPlugin as SearchPlugin
from aioqbt.api.types import SearchPluginCategory as SearchPluginCategory
from aioqbt.api.types import SearchResultEntry as SearchResultEntry
from aioqbt.api.types import SyncCategory as SyncCategory
from aioqbt.api.types import SyncMainData as SyncMainData
from aioqbt.api.types import SyncPeer as SyncPeer
from aioqbt.api.types import SyncServerState as SyncServerState
from aioqbt.api.types import SyncTorrentInfo as SyncTorrentInfo
from aioqbt.api.types import SyncTorrentPeers as SyncTorrentPeers
from aioqbt.api.types import TorrentInfo as TorrentInfo
from aioqbt.api.types import TorrentProperties as TorrentProperties
from aioqbt.api.types import Tracker as Tracker
from aioqbt.api.types import TransferInfo as TransferInfo
from aioqbt.api.types import WebSeed as WebSeed



import itertools
import logging
from datetime import datetime
from pathlib import PurePosixPath
from typing import Any, Dict, NamedTuple, Optional
from xml.etree.ElementTree import Element, SubElement
from xml.etree.ElementTree import tostring as stringify_xml

import pytest
import pytest_asyncio
from aiohttp import web
from helper.lang import retry_assert
from helper.web import temporary_web_server

from aioqbt.api.types import Preferences, RSSArticle, RSSFeed, RSSFolder, RSSRule
from aioqbt.client import APIClient
from aioqbt.version import APIVersion

# Note:
# tests usually create their RSS folders under root folder.
# This attempts to avoid interference from other tests.

NAME_TABLE = {
    "news": ["world", "local", "tech", "business", "health"],
    "elements": [
        f"{a} {b}"
        for a, b in itertools.product(
            ("gaseous", "liquid", "solid"),
            ("hydrogen", "carbon", "oxygen", "sodium"),
        )
    ],
}


def _build_rss_xml(
    title: str,
    count: int,
    base_url: str,
    *,
    season: Optional[int] = None,
    desc: str = "description",
) -> Element:
    """Build RSS in XML tree"""
    rss = Element("rss", {"version": "2.0"})
    channel = SubElement(rss, "channel")

    title_elem = SubElement(channel, "title")
    title_elem.text = title

    desc_elem = SubElement(channel, desc)
    desc_elem.text = desc

    names = NAME_TABLE.get(title)

    for idx in range(count):
        item_id = idx + 1

        item = SubElement(channel, "item")

        if season is None:
            episode = ""
        else:
            episode = f" s{season:02d}e{item_id:02d}"

        if names is None:
            item_name = f"item {item_id}"
        else:
            item_name = names[idx % len(names)]

        item_title = SubElement(item, "title")
        item_title.text = f"{title} - {item_name}{episode}"

        item_link = SubElement(item, "link")
        item_link.text = f"{base_url}/{title}-item-{item_id}"

        item_desc = SubElement(item, "description")
        item_desc.text = f"{title} item {item_id} desc"

    return rss


class FeedService(NamedTuple):
    url: str


@pytest_asyncio.fixture
async def feed_service():
    """
    Download RSS feed by HTTP

    The content is generated based on request path and query.
    """

    async def handler(request: web.BaseRequest):
        path = PurePosixPath(request.path)
        suffix = path.suffix

        try:
            if suffix == ".xml":
                title = str(request.query.get("title", path.stem))
                count = int(request.query.get("count", 1))
                try:
                    season = int(request.query.get("season"))  # type: ignore[arg-type]
                except (ValueError, TypeError):
                    season = None

                # utf-8 encoding result in bytes
                # xml_declaration was added in py38. do it manually later.
                body = stringify_xml(
                    _build_rss_xml(title, count, url, season=season),
                    encoding="utf-8",
                )
                assert isinstance(body, bytes)
                body = b"<?xml version='1.0' encoding='utf-8'?>" + body

                return web.Response(
                    body=body,
                    content_type="text/xml",
                )
        except web.HTTPException:
            # HTTP error response
            raise
        except BaseException:
            logging.error(
                "Exception when handling request %r",
                request,
                exc_info=True,
            )
            raise

        raise web.HTTPNotFound()

    async with temporary_web_server(handler) as url:
        yield FeedService(
            url=url,
        )


async def get_items(client: APIClient, with_data: bool) -> Dict[str, Any]:
    items = await client.request_json(
        "GET",
        "rss/items",
        params={
            "withData": "true" if with_data else "false",
        },
    )
    assert isinstance(items, dict)
    return items


@pytest.mark.asyncio
async def test_basic(client: APIClient, feed_service: FeedService):
    # test:
    # add_folder
    # add_feed
    # move_item
    # remove_item

    url = feed_service.url

    await client.rss.add_folder(r"news")
    await client.rss.add_folder(r"news\world")
    await client.rss.add_folder(r"news\asia")
    await client.rss.add_feed(f"{url}/alpha.xml", r"news\world\alpha")
    await client.rss.add_feed(f"{url}/beta.xml?count=2", r"news\world\beta")
    await client.rss.move_item(r"news\world\beta", r"news\asia\beta")

    def assert_feed(d: object, with_data: bool):
        """assert d is a dict of feed item"""
        assert isinstance(d, dict)
        assert d.keys() >= {"uid", "url"}
        assert isinstance(d["uid"], str)
        assert isinstance(d["url"], str)

        if not with_data:
            return

        assert d.keys() >= {"title", "lastBuildDate", "isLoading", "hasError", "articles"}
        assert isinstance(d["isLoading"], bool)
        assert isinstance(d["hasError"], bool)
        assert isinstance(d["articles"], list)

    async def assert_items(with_data: bool):
        items = await get_items(client, with_data)

        assert isinstance(items, dict)
        assert items.keys() >= {"news"}

        assert isinstance(items["news"], dict)
        assert items["news"].keys() == {"world", "asia"}
        assert items["news"]["world"].keys() == {"alpha"}
        assert items["news"]["asia"].keys() == {"beta"}

        assert_feed(items["news"]["world"]["alpha"], with_data)
        assert_feed(items["news"]["asia"]["beta"], with_data)

    await assert_items(False)
    await assert_items(True)

    await client.rss.remove_item(r"news")
    items = await get_items(client, False)
    assert "news" not in items


@pytest.mark.asyncio
async def test_items(client: APIClient, feed_service: FeedService, temp_prefs):
    # enable rss_processing_enabled to refresh RSS feeds in old qBittorrent versions
    prefs = Preferences()
    prefs["rss_processing_enabled"] = True
    await client.app.set_preferences(prefs)

    name = "items"
    url = feed_service.url

    await client.rss.add_folder(name)
    await client.rss.add_feed(f"{url}/items/news.xml", rf"{name}\news")

    @retry_assert
    async def assert_items(with_data: bool):
        root = await client.rss.items(with_data)

        assert isinstance(root, RSSFolder)

        assert "items" in root
        items = root["items"]
        assert isinstance(items, RSSFolder)

        # RSSFolder is a mapping
        assert len(items) == 1
        assert list(items) == ["news"]

        with pytest.raises(KeyError):
            _ = root["badkey"]

        with pytest.raises(KeyError):
            _ = root[r"\items"]  # empty part

        with pytest.raises(KeyError):
            _ = root[r"items\badkey"]

        assert "news" in items
        assert root[r"items\news"] == items[r"news"]

        feed = items["news"]
        assert isinstance(feed, RSSFeed)

        assert isinstance(feed.url, str)
        assert isinstance(feed.uid, str)

        if with_data:
            assert isinstance(feed.title, str)
            assert isinstance(feed.lastBuildDate, str)
            assert isinstance(feed.isLoading, bool)
            assert isinstance(feed.hasError, bool)
            assert isinstance(feed.articles, list)
            assert len(feed.articles)
            assert set(type(s) for s in feed.articles) == {RSSArticle}

            # check one of articles.
            article = feed.articles[0]

            assert isinstance(article.id, str)
            assert isinstance(article.title, str)
            assert isinstance(article.description, str)
            assert isinstance(article.date, datetime)
            assert isinstance(article.link, str)
            assert isinstance(article.torrentURL, str)

    await assert_items(False)
    await assert_items(True)

    await client.rss.remove_item(name)


@pytest.mark.asyncio
async def test_refresh_items(client: APIClient, feed_service: FeedService):
    if APIVersion.compare(client.api_version, (2, 2, 1)) < 0:
        pytest.skip("API 2.2.1")

    name = "refresh"
    url = feed_service.url

    await client.rss.add_feed(f"{url}/refresh.xml?count=2", name)

    async def assert_newly_added():
        items = await get_items(client, True)
        feed = items[name]
        assert isinstance(feed, dict)
        assert len(feed["articles"]) == 0

    await assert_newly_added()

    await client.rss.refresh_item(name)

    @retry_assert
    async def assert_refresh():
        items = await get_items(client, True)
        feed = items[name]
        assert isinstance(feed, dict)
        assert len(feed["articles"]) == 2

    await assert_refresh()

    await client.rss.remove_item(name)


@pytest.mark.asyncio
async def test_rules(client: APIClient, temp_prefs):
    rule1 = RSSRule(
        enabled=False,
        mustContain="mustContain",
        mustNotContain="mustNotContain",
        episodeFilter="1x1-13;",
    )
    rule2 = RSSRule(mustContain="rule2")
    rule3 = RSSRule(
        mustContain="rule3",
    )

    await client.rss.set_rule("rule1", rule1)
    await client.rss.set_rule("rule2", rule2)
    await client.rss.set_rule("rule3", rule3)
    await client.rss.remove_rule("rule2")
    await client.rss.rename_rule("rule3", "rule3new")

    def assert_rule(rule: RSSRule):
        assert isinstance(rule, dict)
        assert rule.keys() >= {"enabled", "mustContain", "mustNotContain", "useRegex"}

    @retry_assert
    async def assert_rule1_created():
        rules = await client.rss.rules()
        assert isinstance(rules, dict), "bad results"
        assert "rule1" in rules
        assert_rule(rules["rule1"])

    await assert_rule1_created()

    @retry_assert
    async def assert_rule2_removed():
        rules = await client.rss.rules()
        assert "rule2" not in rules

    await assert_rule2_removed()

    @retry_assert
    async def assert_rule3_renamed():
        rules = await client.rss.rules()
        assert "rule3" not in rules
        assert "rule3new" in rules
        assert_rule(rules["rule3new"])

    if APIVersion.compare(client.api_version, (2, 6, 1)) >= 0:
        # rename() is not working before API 2.6.1
        # See qbittorrent commit 1b5dd0aa2d2998934adc0dd1f1eb210dba3a9ae9
        await assert_rule3_renamed()

    await client.rss.remove_rule("rule1")
    await client.rss.remove_rule("rule3new")

    empty = await client.rss.rules()

    assert len(empty) == 0


@pytest.mark.asyncio
async def test_mark_as_read(client: APIClient):
    if APIVersion.compare(client.api_version, (2, 5, 1)) < 0:
        pytest.skip("API 2.5.1")

    ns = "marking"

    # no error will report even if feed does not exist.
    await client.rss.mark_as_read(rf"{ns}\news", "hello-world")
    await client.rss.mark_as_read(rf"{ns}\news")


@pytest.mark.asyncio
async def test_matching(client: APIClient, feed_service: FeedService):
    if APIVersion.compare(client.api_version, (2, 5, 1)) < 0:
        pytest.skip("API 2.5.1")

    ns = "matching"
    elements_url = f"{feed_service.url}/{ns}/elements.xml?count=12"
    news_url = f"{feed_service.url}/{ns}/news.xml?count=10"

    await client.rss.add_folder(ns)
    await client.rss.add_feed(elements_url, rf"{ns}\elements")
    await client.rss.add_feed(news_url, rf"{ns}\news")
    await client.rss.refresh_item(ns)

    rule_name = "matching-elements"
    rule = RSSRule(
        mustContain="gas", mustNotContain="hydrogen", affectedFeeds=[elements_url, news_url]
    )
    await client.rss.set_rule(rule_name, rule)

    @retry_assert
    async def assert_matching():
        elements_result = await client.rss.matching_articles(rule_name)
        assert isinstance(elements_result, dict)
        assert elements_result.keys() == {"elements"}
        assert isinstance(elements_result["elements"], list)
        assert all("gas" in s for s in elements_result["elements"])
        assert not any("hydrogen" in s for s in elements_result["elements"])

    await assert_matching()
    await client.rss.remove_item(ns)
    await client.rss.remove_rule(rule_name)



"""
Types utilized and returned by API methods.
"""
from datetime import datetime, timedelta
from typing import Any, Dict, Iterator, List, Mapping, Optional, Type, TypeVar, Union, overload

from typing_extensions import Final, Literal, TypedDict

from aioqbt._compat import IntEnum, StrEnum
from aioqbt.chrono import Minutes, TimeUnit
from aioqbt.converter import (
    DateTimeConverter,
    DurationConverter,
    EnumConverter,
    RFC2822DateTimeConverter,
    ScalarListConverter,
)
from aioqbt.mapper import declarative, field


# define enums first
class TorrentState(StrEnum):
    """
    Possible torrent states in :attr:`.TorrentInfo.state`.
    """

    ERROR = "error"
    MISSING_FILES = "missingFiles"
    UPLOADING = "uploading"
    PAUSED_UP = "pausedUP"
    QUEUED_UP = "queuedUP"
    STALLED_UP = "stalledUP"
    CHECKING_UP = "checkingUP"
    FORCED_UP = "forcedUP"
    ALLOCATING = "allocating"
    DOWNLOADING = "downloading"
    META_DL = "metaDL"
    PAUSED_DL = "pausedDL"
    QUEUED_DL = "queuedDL"
    STALLED_DL = "stalledDL"
    CHECKING_DL = "checkingDL"
    FORCED_DL = "forcedDL"
    CHECKING_RESUME_DATA = "checkingResumeData"
    MOVING = "moving"
    UNKNOWN = "unknown"


class InfoFilter(StrEnum):
    """
    Torrent state filter in :meth:`.TorrentsAPI.info`.
    """

    ALL = "all"
    DOWNLOADING = "downloading"
    SEEDING = "seeding"
    COMPLETED = "completed"
    RESUMED = "resumed"
    PAUSED = "paused"
    ACTIVE = "active"
    INACTIVE = "inactive"
    STALLED = "stalled"
    STALLED_UPLOADING = "stalled_uploading"
    STALLED_DOWNLOADING = "stalled_downloading"
    ERRORED = "errored"


class PieceState(IntEnum):
    """
    Piece state in :meth:`.TorrentsAPI.piece_states` results.
    """

    UNAVAILABLE = 0
    DOWNLOADING = 1
    DOWNLOADED = 2


class TrackerStatus(IntEnum):
    """
    Tracker status in :attr:`.Tracker.status`.
    """

    DISABLED = 0
    NOT_CONTACTED = 1
    WORKING = 2
    UPDATING = 3
    NOT_WORKING = 4


class RatioLimits(IntEnum):
    """
    Special values of ratio limit.
    """

    UNSET = -1


RatioLimitTypes = Union[float, RatioLimits]


class SeedingTimeLimits(IntEnum):
    """
    Special values of seeding time limit.
    """

    GLOBAL = -2
    UNLIMITED = -1


SeedingTimeLimitTypes = Union[timedelta, Minutes, SeedingTimeLimits]


class InactiveSeedingTimeLimits(IntEnum):
    GLOBAL = -2
    UNLIMITED = -1


InactiveSeedingTimeLimitTypes = Union[timedelta, Minutes, InactiveSeedingTimeLimits]


class StopCondition(StrEnum):
    """
    Stopping condition to pause torrents.
    """

    NONE = "None"
    METADATA_RECEIVED = "MetadataReceived"
    FILES_CHECKED = "FilesChecked"


class ContentLayout(StrEnum):
    """
    Content layout that downloaded files are organized.
    """

    ORIGINAL = "Original"
    SUBFOLDER = "Subfolder"
    NO_SUBFOLDER = "NoSubfolder"


class FilePriority(IntEnum):
    """
    File priority in :meth:`.TorrentsAPI.file_prio` and :attr:`.FileEntry.priority`.
    """

    NO_DOWNLOAD = 0
    NORMAL = 1
    _DEFAULT = 4  # libtorrent defaults to 4, but invalid in bittorrent
    HIGH = 6
    MAXIMAL = 7


class ConnectionStatus(StrEnum):
    """
    Connection status in :attr:`.TransferInfo.connection_status`.
    """

    CONNECTED = "connected"
    FIREWALLED = "firewalled"
    DISCONNECTED = "disconnected"


# values mapped to None
_DATETIME_NONE_TABLE = dict.fromkeys(
    (
        -1,
        0xFFFF_FFFF,  # -1 as u32
    )
)
_DURATION_NONE_TABLE = dict.fromkeys((-1,))
_IE = TypeVar("_IE", bound=IntEnum)
_SE = TypeVar("_SE", bound=StrEnum)


@overload
def _table_from_enum(cls: Type[_IE]) -> Dict[int, _IE]:
    ...


@overload
def _table_from_enum(cls: Type[_SE]) -> Dict[str, _SE]:
    ...


def _table_from_enum(cls: Union[Type[IntEnum], Type[StrEnum]]) -> Dict[Any, Any]:
    if issubclass(cls, IntEnum):
        return {int(s): s for s in cls.__members__.values()}
    elif issubclass(cls, StrEnum):
        return {str(s): s for s in cls.__members__.values()}
    else:
        raise AssertionError("unreachable")


# define dataclasses after enums
@declarative
class BuildInfo:
    """
    See :meth:`.AppAPI.build_info`.
    """

    qt: str
    libtorrent: str
    boost: str
    openssl: str
    zlib: str  # undocumented, found in API v2.5.1
    bitness: int


class Preferences(TypedDict, total=False):
    """
    Dict of preferences.

    .. note::

        Preference keys may be added/changed/removed across versions.
        Please refer to :APIWiki:`the documentation <#get-application-preferences>`.

    """

    locale: str
    performance_warning: bool

    file_log_enabled: bool
    file_log_path: str
    file_log_backup_enabled: bool
    file_log_max_size: int
    file_log_delete_old: bool
    file_log_age: int
    file_log_age_type: int

    torrent_content_layout: str
    add_to_top_of_queue: bool
    create_subfolder_enabled: bool  # removed in v4.3.2
    start_paused_enabled: bool
    torrent_stop_condition: str
    merge_trackers: bool
    auto_delete_mode: int
    preallocate_all: bool
    incomplete_files_ext: bool

    auto_tmm_enabled: bool
    torrent_changed_tmm_enabled: bool
    save_path_changed_tmm_enabled: bool
    category_changed_tmm_enabled: bool
    use_subcategories: bool
    save_path: str
    temp_path_enabled: bool
    temp_path: str
    use_category_paths_in_manual_mode: bool
    export_dir: str
    export_dir_fin: str

    scan_dirs: Dict[str, Union[int, str]]

    excluded_file_names_enabled: bool
    excluded_file_names: str

    mail_notification_enabled: bool
    mail_notification_sender: str
    mail_notification_email: str
    mail_notification_smtp: str
    mail_notification_ssl_enabled: bool
    mail_notification_auth_enabled: bool
    mail_notification_username: str
    mail_notification_password: str

    autorun_on_torrent_added_enabled: bool
    autorun_on_torrent_added_program: str

    autorun_enabled: bool
    autorun_program: str

    listen_port: int
    random_port: bool
    upnp: bool

    max_connec: int
    max_connec_per_torrent: int
    max_uploads: int
    max_uploads_per_torrent: int

    proxy_type: Union[int, str]  # int -> str in v4.6.0
    proxy_ip: str
    proxy_port: int
    proxy_auth_enabled: bool
    proxy_username: str
    proxy_password: str
    proxy_hostname_lookup: bool

    proxy_bittorrent: bool  # found in v4.6.0
    proxy_torrents_only: bool
    proxy_peer_connections: bool
    proxy_rss: bool  # found in v4.6.0
    proxy_misc: bool  # found in v4.6.0

    ip_filter_enabled: bool
    ip_filter_path: str
    ip_filter_trackers: bool
    banned_IPs: str

    dl_limit: int
    up_limit: int
    alt_dl_limit: int
    alt_up_limit: int
    bittorrent_protocol: int
    limit_utp_rate: bool
    limit_tcp_overhead: bool
    limit_lan_peers: bool

    scheduler_enabled: bool
    schedule_from_hour: int
    schedule_from_min: int
    schedule_to_hour: int
    schedule_to_min: int
    scheduler_days: int

    dht: bool
    pex: bool
    lsd: bool
    encryption: int
    anonymous_mode: bool

    max_active_checking_torrents: int

    queueing_enabled: bool
    max_active_downloads: int
    max_active_torrents: int
    max_active_uploads: int
    dont_count_slow_torrents: bool
    slow_torrent_dl_rate_threshold: int
    slow_torrent_ul_rate_threshold: int
    slow_torrent_inactive_timer: int

    max_ratio_enabled: bool
    max_ratio: int
    max_seeding_time_enabled: bool
    max_seeding_time: int
    max_inactive_seeding_time_enabled: bool  # 4.6.0
    max_inactive_seeding_time: int  # 4.6.0
    max_ratio_act: int

    add_trackers_enabled: bool
    add_trackers: str

    web_ui_domain_list: str
    web_ui_address: str
    web_ui_port: int
    web_ui_upnp: bool
    use_https: bool
    web_ui_https_cert_path: str
    web_ui_https_key_path: str

    web_ui_username: str
    bypass_local_auth: bool
    bypass_auth_subnet_whitelist_enabled: bool
    bypass_auth_subnet_whitelist: str
    web_ui_max_auth_fail_count: int
    web_ui_ban_duration: int
    web_ui_session_timeout: int

    alternative_webui_enabled: bool
    alternative_webui_path: str

    web_ui_clickjacking_protection_enabled: bool
    web_ui_csrf_protection_enabled: bool
    web_ui_secure_cookie_enabled: bool
    web_ui_host_header_validation_enabled: bool

    web_ui_use_custom_http_headers_enabled: bool
    web_ui_custom_http_headers: str

    web_ui_reverse_proxy_enabled: bool
    web_ui_reverse_proxies_list: str

    dyndns_enabled: bool
    dyndns_service: int
    dyndns_username: str
    dyndns_password: str
    dyndns_domain: str

    # rss
    rss_refresh_interval: int
    rss_max_articles_per_feed: int
    rss_processing_enabled: bool
    rss_auto_downloading_enabled: bool
    rss_download_repack_proper_episodes: bool
    rss_smart_episode_filters: str

    # advanced
    resume_data_storage_type: str
    memory_working_set_limit: int
    current_network_interface: str
    current_interface_name: str  # v4.6.0
    current_interface_address: str
    save_resume_data_interval: int
    torrent_file_size_limit: int
    recheck_completed_torrents: bool
    refresh_interval: int
    resolve_peer_countries: bool
    reannounce_when_address_changed: bool

    # libtorrent
    bdecode_depth_limit: int
    bdecode_token_limit: int
    async_io_threads: int
    hashing_threads: int
    file_pool_size: int
    checking_memory_use: int
    disk_cache: int
    disk_cache_ttl: int
    disk_queue_size: int
    disk_io_type: int
    disk_io_read_mode: int
    disk_io_write_mode: int
    enable_os_cache: bool  # removed
    enable_coalesce_read_write: bool
    enable_piece_extent_affinity: bool
    enable_upload_suggestions: bool
    send_buffer_watermark: int
    send_buffer_low_watermark: int
    send_buffer_watermark_factor: int
    connection_speed: int
    socket_send_buffer_size: int  # found in v4.6.0
    socket_receive_buffer_size: int  # found in v4.6.0
    socket_backlog_size: int
    outgoing_ports_min: int
    outgoing_ports_max: int
    upnp_lease_duration: int
    peer_tos: int
    utp_tcp_mixed_mode: int
    idn_support_enabled: bool
    enable_multi_connections_from_same_ip: bool
    validate_https_tracker_certificate: bool
    ssrf_mitigation: bool
    block_peers_on_privileged_ports: bool
    enable_embedded_tracker: bool
    embedded_tracker_port: int
    embedded_tracker_port_forwarding: bool
    upload_slots_behavior: int
    upload_choking_algorithm: int
    announce_to_all_trackers: bool
    announce_to_all_tiers: bool
    announce_ip: str
    max_concurrent_http_announces: int
    stop_tracker_timeout: int
    peer_turnover: int
    peer_turnover_cutoff: int
    peer_turnover_interval: int
    request_queue_size: int

    # removed keys
    force_proxy: bool
    ssl_cert: str
    ssl_key: str
    web_ui_password: str


@declarative
class NetworkInterface:
    """
    See :meth:`.AppAPI.network_interface_list`.
    """

    name: str
    value: str


@declarative
class TorrentInfo:
    """
    Obtained from :meth:`.TorrentsAPI.info`.

    Also see :APIWiki:`qBittorrent Wiki <#get-torrent-list>` for attribute meanings.

    .. include:: shared/missing_attributes.rst
    """

    hash: str
    """Torrent ID (info hash)"""

    infohash_v1: str  # API 2.8.4
    infohash_v2: str  # API 2.8.4
    """``infohash_v1`` and ``infohash_v2`` are available since v4.4.0"""

    name: str
    """Torrent name."""

    magnet_uri: str
    size: int
    progress: float
    dlspeed: int
    upspeed: int
    priority: int
    num_seeds: int
    num_complete: int
    num_leechs: int
    num_incomplete: int

    state: Union[str, TorrentState] = field(
        convert=EnumConverter(TorrentState),
    )
    eta: timedelta = field(
        convert=DurationConverter(TimeUnit.SECONDS),
    )
    seq_dl: bool
    f_l_piece_prio: bool

    category: str
    tags: List[str] = field(
        convert=ScalarListConverter(","),
    )
    super_seeding: bool
    force_start: bool
    save_path: str
    download_path: str  # API v2.8.4
    content_path: str  # API v2.6.1
    added_on: datetime = field(
        convert=DateTimeConverter(),  # unix timestamp
    )
    completion_on: datetime = field(
        convert=DateTimeConverter(),  # unix timestamp
    )
    tracker: str
    trackers_count: int
    dl_limit: int
    up_limit: int
    downloaded: int
    uploaded: int
    downloaded_session: int
    uploaded_session: int
    amount_left: int
    completed: int
    max_ratio: float
    max_seeding_time: Optional[timedelta] = field(
        convert=DurationConverter(TimeUnit.MINUTES, _DURATION_NONE_TABLE),
    )
    max_inactive_seeding_time: Optional[timedelta] = field(
        convert=DurationConverter(TimeUnit.MINUTES, _DURATION_NONE_TABLE),
    )
    ratio: float
    ratio_limit: Union[float, RatioLimits]
    seeding_time_limit: Union[timedelta, int, SeedingTimeLimits] = field(
        convert=DurationConverter(TimeUnit.MINUTES, _table_from_enum(SeedingTimeLimits)),
    )
    inactive_seeding_time_limit: Union[timedelta, int, InactiveSeedingTimeLimits] = field(
        convert=DurationConverter(TimeUnit.MINUTES, _table_from_enum(SeedingTimeLimits)),
    )
    seen_complete: datetime = field(
        convert=DateTimeConverter(),
    )
    auto_tmm: bool
    time_active: timedelta = field(
        convert=DurationConverter(TimeUnit.SECONDS),
    )
    seeding_time: timedelta = field(
        convert=DurationConverter(TimeUnit.SECONDS),
    )
    last_activity: datetime = field(
        convert=DateTimeConverter(),
    )
    availability: float

    total_size: int

    def __repr__(self) -> str:
        return f"<{type(self).__name__} {self.hash} {self.state} {self.name!r}>"


@declarative
class TorrentProperties:
    """
    Obtained from :meth:`.TorrentsAPI.properties`.

    .. include:: shared/missing_attributes.rst
    """

    infohash_v1: str  # API v2.8.3
    infohash_v2: str  # API v2.8.3
    """``infohash_v1`` and ``infohash_v2`` are available since v4.4.0"""

    name: str  # API v2.8.19
    hash: str  # API v2.8.19
    """``name`` and ``hash`` are available since v4.5.2"""

    time_elapsed: timedelta = field(
        convert=DurationConverter(TimeUnit.SECONDS),
    )
    seeding_time: timedelta = field(
        convert=DurationConverter(TimeUnit.MINUTES),
    )
    eta: timedelta = field(
        convert=DurationConverter(TimeUnit.SECONDS),
    )
    nb_connections: int
    nb_connections_limit: int
    total_downloaded: int
    total_downloaded_session: int
    total_uploaded: int
    total_uploaded_session: int
    dl_speed: int
    dl_speed_avg: int
    up_speed: int
    up_speed_avg: int
    dl_limit: int
    up_limit: int
    total_wasted: int
    seeds: int
    seeds_total: int
    peers: int
    peers_total: int
    share_ratio: float
    reannounce: timedelta = field(
        convert=DurationConverter(TimeUnit.SECONDS),
    )
    total_size: int
    pieces_num: int
    piece_size: int
    pieces_have: int
    created_by: str
    is_private: bool  # v2.8.20
    addition_date: datetime = field(
        convert=DateTimeConverter(),
    )

    last_seen: Optional[datetime] = field(
        convert=DateTimeConverter(_DATETIME_NONE_TABLE),
    )
    completion_date: Optional[datetime] = field(
        convert=DateTimeConverter(_DATETIME_NONE_TABLE),
    )
    creation_date: Optional[datetime] = field(
        convert=DateTimeConverter(_DATETIME_NONE_TABLE),
    )

    save_path: str
    download_path: str  # v2.8.4
    comment: str

    def __repr__(self) -> str:
        cls_name = type(self).__name__

        name = getattr(self, "name", None)
        hash = getattr(self, "hash", None)

        if name is None and hash is None:
            return f"<{cls_name} at 0x{hex(id(self))}>"
        else:
            return f"<{cls_name} hash={hash!s} name={name!r}>"


@declarative
class Tracker:
    """
    See :meth:`.TorrentsAPI.trackers`.
    """

    url: str
    status: Union[int, TrackerStatus] = field(
        convert=EnumConverter(TrackerStatus),
    )
    tier: int
    num_peers: int
    num_seeds: int
    num_leeches: int
    num_downloaded: int
    msg: str

    def __repr__(self) -> str:
        cls_name = type(self).__name__
        title = self.url or ""
        msg = f" {self.msg!r}" if self.msg else ""

        if title.startswith("** [") and title.endswith("] **"):
            return f"<{cls_name} {title[4:-4]}{msg}>"
        else:
            return f"<{cls_name} {title!r}{msg}>"

    def is_special(self) -> bool:
        url = self.url or ""
        return url.startswith("** [") and url.endswith("] **")


@declarative
class WebSeed:
    """
    See :meth:`.TorrentsAPI.webseeds`.
    """

    url: str


@declarative
class FileEntry:
    """
    See :meth:`.TorrentsAPI.files`.
    """

    name: str
    size: int
    progress: float
    priority: Union[int, FilePriority]
    piece_range: List[int]
    is_seed: bool = field(repr=False)  # only available in the first item
    availability: float
    index: int = field(repr=False)  # since API v2.8.2


@declarative
class Category:
    """
    See :meth:`.TorrentsAPI.categories`.
    """

    name: str
    savePath: str


@declarative
class LogMessage:
    """
    See :meth:`.LogAPI.main`.
    """

    id: int
    message: str
    timestamp: int
    type: int


@declarative
class LogPeer:
    """
    See :meth:`.LogAPI.peers`.
    """

    id: int
    ip: str
    timestamp: int
    blocked: bool
    reason: str


class SyncTorrentInfo(TypedDict, total=False):
    """
    Dict of torrent info in :attr:`.SyncMainData.torrents`.
    """

    # This type and TorrentInfo share mostly identical set of keys
    # Annotations are more primordial here.

    # Due to difference algorithm, "hash" is always filtered out.
    # hash: str

    infohash_v1: str
    infohash_v2: str

    name: str
    size: int

    magnet_uri: str
    progress: float
    dlspeed: int
    upspeed: int
    priority: int
    num_seeds: int
    num_complete: int
    num_leechs: int
    num_incomplete: int

    state: str
    eta: int
    seq_dl: bool
    f_l_piece_prio: bool

    category: str
    tags: str
    super_seeding: bool
    force_start: bool
    save_path: str
    download_path: str
    content_path: str
    added_on: int
    completion_on: int
    tracker: str
    trackers_count: int
    dl_limit: int
    up_limit: int
    downloaded: int
    uploaded: int
    downloaded_session: int
    uploaded_session: int
    amount_left: int
    completed: int
    max_ratio: float
    max_seeding_time: int
    max_inactive_seeding_time: int
    ratio: float
    ratio_limit: float
    seeding_time_limit: int
    inactive_seeding_time_limit: int
    seen_complete: int
    auto_tmm: bool
    time_active: int
    seeding_time: int
    last_activity: int
    availability: float

    total_size: int


class SyncCategory(TypedDict, total=False):
    """
    Dict of category properties in :attr:`.SyncMainData.categories`.
    """

    name: str
    savePath: str


class SyncServerState(TypedDict, total=False):
    """
    Dict of qBittorrent status and statistics in :attr:`.SyncMainData.server_state`.
    """

    connection_status: Union[str, ConnectionStatus]
    dht_nodes: int
    dl_info_data: int
    dl_info_speed: int
    dl_rate_limit: int
    up_info_data: int
    up_info_speed: int
    up_rate_limit: int

    alltime_dl: int
    alltime_ul: int
    total_wasted_session: int
    global_ratio: str
    total_peer_connections: int

    queueing: bool
    use_alt_speed_limits: bool
    refresh_interval: int
    free_space_on_disk: int
    use_subcategories: bool

    average_time_queue: int
    read_cache_hits: str
    read_cache_overload: str
    write_cache_overload: str
    queued_io_jobs: int
    total_buffers_size: int
    total_queued_size: int


@declarative
class SyncMainData:
    """
    Sync results obtained from :meth:`.SyncAPI.maindata`.
    """

    rid: int
    full_update: bool = field(
        default=False,
    )

    torrents: Dict[str, SyncTorrentInfo] = field(
        default_factory=dict,
    )
    torrents_removed: List[str] = field(
        default_factory=list,
    )

    categories: Dict[str, SyncCategory] = field(
        default_factory=dict,
    )
    categories_removed: List[str] = field(
        default_factory=list,
    )

    tags: List[str] = field(
        default_factory=list,
    )
    tags_removed: List[str] = field(
        default_factory=list,
    )

    # trackers are new in 4.6.0
    trackers: Dict[str, List[str]] = field(
        default_factory=dict,
    )
    trackers_removed: List[str] = field(
        default_factory=list,
    )

    server_state: SyncServerState = field(
        default_factory=lambda: SyncServerState(),
    )


class SyncPeer(TypedDict, total=False):
    """
    Dict of peer info in :attr:`.SyncTorrentPeers.peers`.
    """

    ip: str
    port: int
    client: str
    progress: float
    dl_speed: int
    up_speed: int
    downloaded: int
    uploaded: int
    connection: str
    flags: str
    flags_desc: str
    relevance: float
    files: str
    country_code: str
    country: str


@declarative
class SyncTorrentPeers:
    """
    See :meth:`.SyncAPI.torrent_peers`.
    """

    rid: int
    full_update: bool = field(default=False)
    # "show_flags" may be true, false or missing
    show_flags: Optional[bool] = field(default=None)
    peers: Dict[str, SyncPeer] = field(default_factory=dict)


@declarative
class TransferInfo:
    """
    See :meth:`.TransferAPI.info`.
    """

    dl_info_speed: int
    dl_info_data: int
    up_info_speed: int
    up_info_data: int
    dl_rate_limit: int
    up_rate_limit: int
    dht_nodes: int

    connection_status: Union[str, ConnectionStatus] = field(
        convert=EnumConverter(ConnectionStatus),
    )


# RSS related
_RSS_SEPARATOR: Final = "\\"
_FEED_FOLDER_UNION = Union["RSSFeed", "RSSFolder"]


@declarative
class RSSArticle:
    """RSS article."""

    id: str
    title: str
    description: str
    date: datetime = field(
        convert=RFC2822DateTimeConverter(),
    )
    link: str
    torrentURL: str


@declarative
class RSSItem:
    """Base class of :class:`RSSFolder` and :class:`RSSFeed`"""


@declarative
class RSSFolder(RSSItem, Mapping[str, _FEED_FOLDER_UNION]):
    """
    RSSFolder is a container in hierarchical tree.

    .. code-block:: text

        folder
        |-- linux
        |-- news
        |   |-- local
        |   |-- world

    RSSFolder is a dict-like object that children :class:`RSSFeed` and :class:`RSSFolder` are
    accessed by their names: ``folder["linux"]``.
    The number of direct children is returned by ``len(folder)``
    while the names of them by ``folder.keys()``.

    Further children can be accessed by joining names with backslashes.
    The following lines are equivalent::

        folder["news"]["local"]
        folder[r"news\\local"]

    """

    _items: Mapping[str, _FEED_FOLDER_UNION]

    def __iter__(self) -> Iterator[str]:
        return iter(self._items)

    def __len__(self) -> int:
        return len(self._items)

    def __getitem__(self, path: str) -> _FEED_FOLDER_UNION:
        if _RSS_SEPARATOR not in path:
            return self._items[path]

        # use Any to pass typecheck
        items: Mapping[str, _FEED_FOLDER_UNION] = self._items
        parts = path.split(_RSS_SEPARATOR)
        end = len(parts) - 1
        for idx, name in enumerate(parts):
            if not name:
                raise KeyError(f"Empty segment at {idx}: {path!r}")

            try:
                result = items[name]
            except KeyError:
                raise KeyError(f"No item at {idx}: {path!r}")

            if idx == end:
                return result

            if not isinstance(result, RSSFolder):
                raise KeyError(f"Expect RSSFolder at {idx}: {path!r}")

            items = result._items

        raise AssertionError("unreachable")


@declarative
class RSSFeed(RSSItem):
    """
    RSS feed returned from :meth:`.RSSAPI.items`.

    Attributes ``url`` and ``uid``  are always available.

    The other attributes are available if ``with_data=True`` is passed to :meth:`.RSSAPI.items`.
    Otherwise, :class:`AttributeError` raises when accessed.

    Use :func:`hasattr` to check if doubted.
    """

    url: str
    uid: str

    # following attributes are available if with_data is set
    title: str
    lastBuildDate: str
    isLoading: bool
    hasError: bool
    articles: List[RSSArticle]  # entries are processed in RSSAPI.items()

    def __repr__(self) -> str:
        if hasattr(self, "title"):
            return f"<RSSFeed title={self.title!r} url={self.url!r}>"
        else:
            return f"<RSSFeed uid={self.uid!r} url={self.url!r}>"

    __str__ = __repr__


class RSSRule(TypedDict, total=False):
    """
    RSS rule configuration dict.

    Rule dict is returned from :meth:`.RSSAPI.rules`.
    It can be passed as argument to :meth:`.RSSAPI.set_rule`.
    """

    enabled: bool
    priority: int

    useRegex: bool
    mustContain: str
    mustNotContain: str
    episodeFilter: str
    affectedFeeds: List[str]
    savePath: str
    assignedCategory: str
    lastMatch: str
    ignoreDays: int
    addPaused: Optional[bool]
    torrentContentLayout: Optional[str]
    smartFilter: bool
    previouslyMatchedEpisodes: List[str]

    torrentParams: Dict[str, Any]  # TODO update the type when upstream work done


# Search related


@declarative
class SearchJobStart:
    """
    Result of :meth:`.SearchAPI.start`.
    """

    id: int


@declarative
class SearchJobStatus:
    """
    Search job status.
    """

    id: int
    status: Union[str, Literal["Running"], Literal["Stopped"]]
    total: int


@declarative
class SearchResultEntry:
    """Search result entry."""

    fileName: str
    fileUrl: str
    fileSize: str
    nbSeeders: int
    nbLeechers: int
    siteUrl: str


@declarative
class SearchJobResults:
    """Search job results."""

    status: Union[str, Literal["Running"], Literal["Stopped"]]
    results: List[SearchResultEntry]
    total: int


@declarative
class SearchPluginCategory:
    """Category supported by plugin."""

    id: str
    category: str


@declarative
class SearchPlugin:
    """
    Search plugin information.
    """

    enabled: bool
    fullName: str
    name: str
    supportedCategories: Union[List[SearchPluginCategory], List[str]]
    """
    A list of supported categories.

    In qBittorrent 4.3.x and later, this attribute is a list of :class:`.SearchPluginCategory`;
    in earlier versions, a list of localized strings.
    """
    url: str
    version: str



"""
Exceptions raised in :mod:`aioqbt`.

"""
import logging
import sys
from typing import Optional

import aiohttp
from typing_extensions import Self


class AQError(Exception):
    """
    Generic error class.
    """


class MapperError(AQError):
    """
    Raised when mapper operations failed.
    """


class VersionError(AQError):
    """
    Version check failed.
    """


class APIError(AQError):
    """
    Base class of API errors.
    """

    message: str
    """Error message or HTTP reason phrase"""

    status: int
    """HTTP status"""

    resp: Optional[aiohttp.ClientResponse]
    """The closed response object"""

    def __init__(
        self,
        message: str = "",
        status: int = 0,
        resp: Optional[aiohttp.ClientResponse] = None,
    ):
        self.message = message
        self.status = status
        self.resp = resp

    @classmethod
    def from_response(
        cls,
        resp: aiohttp.ClientResponse,
        message: str = "",
    ) -> Self:
        """Create an exception instance based on a response."""

        return cls(
            message=message or str(resp.reason),
            status=resp.status,
            resp=resp,
        )

    def __repr__(self) -> str:
        return f"{type(self).__name__}(status={self.status!r}, message={self.message!r})"

    def __str__(self) -> str:
        return repr(self)


class LoginError(APIError):
    """
    Login has failed.

    This is raised by :meth:`.AuthAPI.login` and HTTP status is 200.
    """


class AddTorrentError(APIError):
    """
    No **new** torrents were added.

    This is raised by :meth:`.TorrentsAPI.add` and HTTP status is 200.
    """


class BadRequestError(APIError):
    """
    Bad request.

    The error is usually raised because of missing or invalid parameters.
    This may be caused by empty value sometimes.

    HTTP status is usually 400.
    """


class ForbiddenError(APIError):
    """
    Forbidden.

    The request to resources is explicitly denied due to permission.

    HTTP status is usually 403.
    """


class NotFoundError(APIError):
    """
    Not found.

    It is likely that the API endpoint is misspelled or qBittorrent need an update.

    HTTP status is usually 404.
    """


class ConflictError(APIError):
    """
    Conflict.

    HTTP status is usually 409.
    """


class UnsupportedMediaTypeError(APIError):
    """
    Unsupported media type.

    HTTP status is usually 415.
    """


# lookup error class by status
# {status: exc_class}
_ERROR_TABLE = {
    400: BadRequestError,
    403: ForbiddenError,
    404: NotFoundError,
    409: ConflictError,
    415: UnsupportedMediaTypeError,
}


def _add_note(
    ex: Optional[BaseException],
    note: str,
    *,
    logger: Optional[logging.Logger] = None,
    log_level: Optional[int] = None,
) -> None:
    """
    Add a note to exception if Python version >= 3.11 or record with logger
    """

    if sys.version_info >= (3, 11):
        ex.add_note(note)
        return

    # fallback to logging
    if logger is None:
        logger = logging.getLogger(f"{__name__}.note")

    if log_level is None:
        log_level = logging.WARNING

    logger.log(log_level, note, exc_info=ex)



from enum import Enum
from typing import NewType

Seconds = NewType("Seconds", int)
Minutes = NewType("Minutes", int)


class TimeUnit(Enum):
    """
    TimeUnit enum.

    Conversion helper between time units
    """

    NANOSECONDS = 1
    MICROSECONDS = 1_000
    MILLISECONDS = 1_000_000
    SECONDS = 1_000_000_000
    MINUTES = 1_000_000_000 * 60
    HOURS = 1_000_000_000 * 3600
    DAYS = 1_000_000_000 * 86400

    def from_duration(self, d: float, unit: "TimeUnit") -> float:
        """Convert value from the given unit to the self unit (float)"""
        return d * unit._value_ / self._value_

    def from_duration_int(self, d: int, unit: "TimeUnit") -> int:
        """Convert value from the given unit to the self unit (int)"""
        return d * unit._value_ // self._value_

    def from_nanos(self, d: float) -> float:
        """Convert nanoseconds to self unit"""
        return d * TimeUnit.NANOSECONDS._value_ / self._value_

    def from_micros(self, d: float) -> float:
        """Convert microseconds to self unit"""
        return d * TimeUnit.MICROSECONDS._value_ / self._value_

    def from_millis(self, d: float) -> float:
        """Convert milliseconds to self unit"""
        return d * TimeUnit.MILLISECONDS._value_ / self._value_

    def from_seconds(self, d: float) -> float:
        """Convert seconds to self unit"""
        return d * TimeUnit.SECONDS._value_ / self._value_

    def from_minutes(self, d: float) -> float:
        """Convert minutes to self unit"""
        return d * TimeUnit.MINUTES._value_ / self._value_

    def from_hours(self, d: float) -> float:
        """Convert hours to self unit"""
        return d * TimeUnit.HOURS._value_ / self._value_

    def from_days(self, d: float) -> float:
        """Convert days to self unit"""
        return d * TimeUnit.DAYS._value_ / self._value_

    @classmethod
    def convert(cls, d: float, src: "TimeUnit", dst: "TimeUnit") -> float:
        """Convert a numeric duration in some unit to one in another."""
        return dst.from_duration(d, src)



import dataclasses
import os
from datetime import timedelta
from typing import TYPE_CHECKING, Any, Dict, Iterable, List, Optional, Tuple, Union, overload

import aiohttp
from typing_extensions import Self

from aioqbt import exc
from aioqbt._decorator import copy_self
from aioqbt._paramdict import ParamDict
from aioqbt.api.types import (
    Category,
    ContentLayout,
    FileEntry,
    InfoFilter,
    RatioLimits,
    SeedingTimeLimits,
    StopCondition,
    TorrentInfo,
    TorrentProperties,
    Tracker,
    WebSeed,
)
from aioqbt.bittorrent import InfoHash, InfoHashes, InfoHashesOrAll, _info_hash_str
from aioqbt.chrono import Minutes, TimeUnit
from aioqbt.client import APIClient, APIGroup, since, virtual
from aioqbt.typing import StrPath
from aioqbt.version import APIVersion, ClientVersion

__all__ = (
    "AddFormBuilder",
    "TorrentsAPI",
)


def _check_iterable_except_str(param: str, value: Iterable[Any]) -> None:
    """Explicitly reject ``str`` as ``Iterable[str]``"""
    if isinstance(value, str):  # pragma: no cover
        raise ValueError(f"{param!r} refused str as iterable")


class TorrentsAPI(APIGroup):
    """
    API methods under ``torrents``.
    """

    async def count(self) -> int:
        """
        Get the number of torrents
        """
        # new in v4.6.1
        result = await self._request_text(
            "GET",
            "torrents/count",
        )
        return int(result)

    async def info(
        self,
        filter: Optional[str] = None,
        category: Optional[str] = None,
        sort: Optional[str] = None,
        reverse: Optional[bool] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        hashes: Optional[InfoHashesOrAll] = None,
        tag: Optional[str] = None,  # since API v2.8.3
    ) -> List[TorrentInfo]:
        """
        Get a list of :class:`.TorrentInfo`.

        To obtain a list of completed torrents sorted by name::

            torrents = await client.torrents.info(
                filter=InfoFilter.COMPLETED,
                sort="name",
            )

        See also :APIWiki:`torrents/info <#get-torrent-list>`
        for filter and result meanings.

        :param filter: State filter: :class:`.InfoFilter` or ``str``.
        :param category: category filter.
        :param sort: Sort results by an attribute/field.
        :param reverse: Reverse the results.
        :param limit: Maximum number of returned results.
        :param offset: Results starting from the ``offset``-th torrents.
        :param hashes: A list of info hashes, or a str ``all``.
        :param tag: Tag filter.
        """
        if isinstance(filter, InfoFilter):
            filter = str(filter)

        if hashes is None:
            params = ParamDict()
        else:
            params = ParamDict.with_hashes(hashes)

        params.optional_str("filter", filter)
        params.optional_str("category", category)
        params.optional_str("sort", sort)
        params.optional_bool("reverse", reverse)
        params.optional_int("limit", limit)
        params.optional_int("offset", offset)

        if tag is not None:
            # API 2.8.3
            params.optional_str("tag", tag)

        return await self._request_mapped_list(
            TorrentInfo,
            "GET",
            "torrents/info",
            params=params,
        )

    async def properties(self, hash: InfoHash) -> TorrentProperties:
        """Get properties of a torrent."""

        props = await self._request_mapped_object(
            TorrentProperties,
            "GET",
            "torrents/properties",
            params=ParamDict.with_hash(hash),
        )

        if not hasattr(props, "hash"):
            props.hash = _info_hash_str(hash)

        return props

    async def trackers(self, hash: InfoHash) -> List[Tracker]:
        """Trackers in a torrent."""

        # Tracker's status may be a string or int, API v2.2.0
        return await self._request_mapped_list(
            Tracker,
            "GET",
            "torrents/trackers",
            params=ParamDict.with_hash(hash),
        )

    async def webseeds(self, hash: InfoHash) -> List[WebSeed]:
        """Web seeds in a torrent."""
        return await self._request_mapped_list(
            WebSeed,
            "GET",
            "torrents/webseeds",
            params=ParamDict.with_hash(hash),
        )

    async def files(
        self,
        hash: InfoHash,
        indexes: Optional[Iterable[int]] = None,
    ) -> List[FileEntry]:
        """Files in a torrent."""
        params = ParamDict.with_hash(hash)

        if indexes is not None:
            # API 2.8.2
            params.optional_list("indexes", indexes, "|")

        return await self._request_mapped_list(
            FileEntry,
            "GET",
            "torrents/files",
            params=params,
        )

    async def piece_states(self, hash: InfoHash) -> List[int]:
        """
        A list of piece states in a torrent.

        To compare results, use following constants from :class:`~.PieceState` enum:

        * :attr:`.PieceState.UNAVAILABLE`
        * :attr:`.PieceState.DOWNLOADING`
        * :attr:`.PieceState.DOWNLOADED`

        """
        return await self._request_json(  # type: ignore[no-any-return]
            "GET",
            "torrents/pieceStates",
            params=ParamDict.with_hash(hash),
        )

    async def piece_hashes(self, hash: InfoHash) -> List[str]:
        """
        A list of piece hashes in a torrent.
        """

        return await self._request_json(  # type: ignore[no-any-return]
            "GET",
            "torrents/pieceHashes",
            params=ParamDict.with_hash(hash),
        )

    async def pause(self, hashes: InfoHashesOrAll) -> None:
        """
        Pause torrents.

        Torrents can be specified by their info hashes.
        Passing ``all`` pauses all torrents.

        """

        await self._request_text(
            "POST",
            "torrents/pause",
            data=ParamDict.with_hashes_or_all(hashes),
        )

    async def resume(self, hashes: InfoHashesOrAll) -> None:
        """
        Resume torrents.

        Torrents can be specified by their info hashes.
        Passing ``all`` resumes all torrents.

        """

        await self._request_text(
            "POST",
            "torrents/resume",
            data=ParamDict.with_hashes_or_all(hashes),
        )

    async def delete(self, hashes: InfoHashesOrAll, delete_files: bool) -> None:
        """
        Delete torrents.

        Torrents can be specified by their info hashes.
        Passing ``all`` deletes all torrents.

        Pass ``True`` to ``delete_files`` to remove downloaded content.
        """

        if hashes != "all":
            _check_iterable_except_str("hashes", hashes)

        data = ParamDict.with_hashes_or_all(hashes)
        data.required_bool("deleteFiles", delete_files)  # default to False

        await self._request_text(
            "POST",
            "torrents/delete",
            data=data,
        )

    async def recheck(self, hashes: InfoHashesOrAll) -> None:
        """Recheck torrents."""

        await self._request_text(
            "POST",
            "torrents/recheck",
            data=ParamDict.with_hashes_or_all(hashes),
        )

    async def reannounce(self, hashes: InfoHashesOrAll) -> None:
        """Reannounce torrents."""

        # since API v2.0.2

        await self._request_text(
            "POST",
            "torrents/reannounce",
            data=ParamDict.with_hashes_or_all(hashes),
        )

    async def add(self, form: aiohttp.FormData) -> None:
        """
        Add torrents by URLs, info hashes, and/or file blobs.

        See :class:`.AddFormBuilder` on how to configure and build
        :class:`~aiohttp.FormData` to submit.

        .. note::

            :exc:`~.exc.AddTorrentError` may raise if no *new* torrents
            are added.

        :param form: form data to submit.
        """

        resp = await self._request(
            "POST",
            "torrents/add",
            data=form,
        )

        async with resp:
            body = await resp.read()

            if body != b"Ok.":
                ex = exc.AddTorrentError.from_response(resp)
                ex.message = body.decode("utf-8")
                raise ex

    async def add_trackers(self, hash: InfoHash, trackers: Iterable[str]) -> None:
        _check_iterable_except_str("trackers", trackers)

        data = ParamDict.with_hash(hash)
        data.required_list("urls", trackers, "\n")

        await self._request_text(
            "POST",
            "torrents/addTrackers",
            data=data,
        )

    @since((2, 2, 0))
    async def edit_tracker(
        self,
        hash: InfoHash,
        orig_url: str,
        new_url: str,
    ) -> None:
        data = ParamDict.with_hash(hash)
        data.required_str("origUrl", orig_url)
        data.required_str("newUrl", new_url)

        await self._request_text(
            "POST",
            "torrents/editTracker",
            data=data,
        )

    async def remove_trackers(self, hash: InfoHash, urls: Iterable[str]) -> None:
        _check_iterable_except_str("urls", urls)

        # Since API v2.2.0
        data = ParamDict.with_hash(hash)
        data.required_list("urls", urls, "|")

        await self._request_text(
            "POST",
            "torrents/removeTrackers",
            data=data,
        )

    async def add_peers(self, hashes: InfoHashes, peers: Iterable[str]) -> None:
        _check_iterable_except_str("peers", peers)

        data = ParamDict.with_hashes(hashes)
        data.required_list("peers", peers, "|")

        await self._request_text(
            "POST",
            "torrents/addPeers",
            data=data,
        )

    async def top_prio(self, hashes: InfoHashesOrAll) -> None:
        await self._request_text(
            "POST",
            "torrents/topPrio",
            data=ParamDict.with_hashes_or_all(hashes),
        )

    async def bottom_prio(self, hashes: InfoHashesOrAll) -> None:
        await self._request_text(
            "POST",
            "torrents/bottomPrio",
            data=ParamDict.with_hashes_or_all(hashes),
        )

    async def increase_prio(self, hashes: InfoHashesOrAll) -> None:
        await self._request_text(
            "POST",
            "torrents/increasePrio",
            data=ParamDict.with_hashes_or_all(hashes),
        )

    async def decrease_prio(self, hashes: InfoHashesOrAll) -> None:
        await self._request_text(
            "POST",
            "torrents/decreasePrio",
            data=ParamDict.with_hashes_or_all(hashes),
        )

    async def file_prio(self, hash: InfoHash, id: Iterable[int], priority: int) -> None:
        """
        Prioritize files in a torrent.

        :param hash: Info hash
        :param id: A list of file indices to prioritize.
        :param priority: Priority, :class:`.FilePriority`.
        """

        # id may be a list since API v2.2.0
        if isinstance(id, int):
            id = (id,)
        data = ParamDict.with_hash(hash)
        data.required_list("id", id, "|")
        data.required_int("priority", priority)

        await self._request_text(
            "POST",
            "torrents/filePrio",
            data=data,
        )

    async def download_limit(self, hashes: InfoHashesOrAll) -> Dict[str, int]:
        """
        Get torrent download limits.

        The result is a dict mapping info hash to download speed limit
        in bytes/second.

        :param hashes: A list of info hashes or ``all`` for all torrents.
        """
        params = ParamDict.with_hashes_or_all(hashes)
        result = await self._request_json(
            "GET",
            "torrents/downloadLimit",
            params=params,
        )
        return result  # type: ignore[no-any-return]

    async def set_download_limit(self, hashes: InfoHashesOrAll, limit: int) -> None:
        """
        Update torrent download limits.

        :param hashes: A list of info hashes or ``all`` for all torrents.
        :param limit: Download limit in bytes/second.
        """

        data = ParamDict.with_hashes_or_all(hashes)
        data.required_int("limit", limit)

        await self._request_text(
            "POST",
            "torrents/setDownloadLimit",
            data=data,
        )

    async def set_share_limits(
        self,
        hashes: InfoHashesOrAll,
        ratio_limit: Union[float, RatioLimits],
        seeding_time_limit: Union[timedelta, int, SeedingTimeLimits],
        inactive_seeding_time_limit: Union[timedelta, int, SeedingTimeLimits, None] = None,
    ) -> None:
        """
        Set share limits for torrents.

        :param hashes: A list of info hashes or ``all`` for all torrents.
        :param ratio_limit: A number or :attr:`.RatioLimits.UNSET`.
        :param seeding_time_limit: :class:`~datetime.timedelta`, or
                :class:`.SeedingTimeLimits` constants.
        :param inactive_seeding_time_limit: :class:`~datetime.timedelta`, or
                :class:`.InactiveSeedingTimeLimits` constants.
                Required since qBittorrent v4.6.0 (API 2.9.2).
        """
        # since API v2.0.1
        client = self._client()

        data = ParamDict.with_hashes_or_all(hashes)
        data.required_float("ratioLimit", ratio_limit)
        data.required_duration("seedingTimeLimit", seeding_time_limit, TimeUnit.MINUTES)

        if inactive_seeding_time_limit is not None:
            data.required_duration(
                "inactiveSeedingTimeLimit", inactive_seeding_time_limit, TimeUnit.MINUTES
            )

        try:
            await client.request_text(
                "POST",
                "torrents/setShareLimits",
                data=data,
            )
        except exc.BadRequestError as ex:
            if (
                inactive_seeding_time_limit is None
                and APIVersion.compare(client.api_version, (2, 9, 2)) >= 0
            ):
                note = "Argument 'inactive_seeding_time_limit' is required since qBittorrent 4.6.0"
                exc._add_note(ex, note, logger=client._logger)

            raise

    async def upload_limit(self, hashes: InfoHashesOrAll) -> Dict[str, int]:
        """
        Get torrent upload limits.

        The result is a dict mapping info hash to upload speed limit
        in bytes/second.

        :param hashes: A list of info hashes or ``all`` for all torrents.
        """
        params = ParamDict.with_hashes_or_all(hashes)

        return await self._request_json(  # type: ignore[no-any-return]
            "GET",
            "torrents/uploadLimit",
            params=params,
        )

    async def set_upload_limit(self, hashes: InfoHashesOrAll, limit: int) -> None:
        """
        Update torrent upload limits.

        :param hashes: A list of info hashes or ``all`` for all torrents.
        :param limit: Upload limit in bytes/second.
        """
        data = ParamDict.with_hashes_or_all(hashes)
        data.required_int("limit", limit)

        await self._request_text(
            "POST",
            "torrents/setUploadLimit",
            data=data,
        )

    async def set_location(
        self,
        hashes: InfoHashesOrAll,
        location: StrPath,
    ) -> None:
        """
        Change location (save path) for torrents.

        This method also turns off auto torrent management (AutoTMM)
        for torrents.

        See also :meth:`~.set_save_path`.

        :param hashes: A list of info hashes or ``all`` for all torrents.
        :param location: Location.
        """
        data = ParamDict.with_hashes_or_all(hashes)
        data.required_path("location", location)

        await self._request_text(
            "POST",
            "torrents/setLocation",
            data=data,
        )

    async def set_save_path(
        self,
        id: InfoHashesOrAll,
        path: StrPath,
    ) -> None:
        """
        Change save path (location) for torrents.

        This method causes no effect to torrents with auto torrent
        management (AutoTMM) enabled.

        Available since qBittorrent v4.4.0.

        See also :meth:`~.set_location`.

        :param id: A list of info hashes or ``all`` for all torrents.
        :param path: Save path.
        """
        # since API v2.8.4
        data = ParamDict.with_hashes_or_all(id, key="id")
        data.required_path("path", path)

        await self._request_text(
            "POST",
            "torrents/setSavePath",
            data=data,
        )

    async def set_download_path(
        self,
        id: InfoHashesOrAll,
        path: StrPath,
    ) -> None:
        """
        Change download path for torrents.

        Available since qBittorrent v4.4.0.

        :param id: A list of info hashes or ``all`` for all torrents.
        :param path: Download path.
        """
        # since API v2.8.4
        data = ParamDict.with_hashes_or_all(id, key="id")
        data.required_path("path", path)

        await self._request_text(
            "POST",
            "torrents/setDownloadPath",
            data=data,
        )

    async def rename(self, hash: InfoHash, name: str) -> None:
        """Rename a torrent."""
        data = ParamDict.with_hash(hash)
        data.required_str("name", name)

        await self._request_text(
            "POST",
            "torrents/rename",
            data=data,
        )

    async def set_category(self, hashes: InfoHashesOrAll, category: str) -> None:
        """
        Change torrents' category.

        :param hashes: A list of info hashes or ``all`` for all torrents.
        :param category: Category name. An empty string indicates no category.
        """

        params = ParamDict.with_hashes_or_all(hashes)
        params.required_str("category", category)

        await self._request_text(
            "POST",
            "torrents/setCategory",
            data=params,
        )

    # since API v2.1.1
    async def categories(self) -> Dict[str, Category]:
        """
        Get categories.

        A dict mapping category name to :class:`.Category` is returned.
        """
        return await self._request_mapped_dict(
            Category,
            "GET",
            "torrents/categories",
        )

    async def create_category(self, category: str, save_path: StrPath) -> None:
        """Create category."""
        data = ParamDict()
        data.required_str("category", category)
        data.required_path("savePath", save_path)

        await self._request_text(
            "POST",
            "torrents/createCategory",
            data=data,
        )

    async def edit_category(self, category: str, save_path: StrPath) -> None:
        """Edit category."""
        # since API v2.1.0
        # empty save_path ("") is default save path

        data = ParamDict()
        data.required_str("category", category)
        data.required_path("savePath", save_path)

        await self._request_text(
            "POST",
            "torrents/editCategory",
            data=data,
        )

    async def remove_categories(self, categories: Iterable[str]) -> None:
        """Remove category."""
        _check_iterable_except_str("categories", categories)

        data = ParamDict()
        data.required_list("categories", categories, "\n")

        await self._request_text(
            "POST",
            "torrents/removeCategories",
            data=data,
        )

    @since((2, 3, 0))
    async def add_tags(self, hashes: InfoHashesOrAll, tags: Iterable[str]) -> None:
        _check_iterable_except_str("tags", tags)

        data = ParamDict.with_hashes_or_all(hashes)
        data.required_list("tags", tags, ",")

        await self._request_text(
            "POST",
            "torrents/addTags",
            data=data,
        )

    @since((2, 3, 0))
    async def remove_tags(self, hashes: InfoHashesOrAll, tags: Iterable[str]) -> None:
        _check_iterable_except_str("tags", tags)

        data = ParamDict.with_hashes_or_all(hashes)
        data.required_list("tags", tags, ",")
        await self._request_text(
            "POST",
            "torrents/removeTags",
            data=data,
        )

    @since((2, 3, 0))
    async def tags(self) -> List[str]:
        return await self._request_json(  # type: ignore[no-any-return]
            "GET",
            "torrents/tags",
        )

    @since((2, 3, 0))
    async def create_tags(self, tags: Iterable[str]) -> None:
        _check_iterable_except_str("tags", tags)

        data = ParamDict()
        data.required_list("tags", tags, ",")
        await self._request_text(
            "POST",
            "torrents/createTags",
            data=data,
        )

    @since((2, 3, 0))
    async def delete_tags(self, tags: Iterable[str]) -> None:
        _check_iterable_except_str("tags", tags)

        data = ParamDict()
        data.required_list("tags", tags, ",")
        await self._request_text(
            "POST",
            "torrents/deleteTags",
            data=data,
        )

    async def set_auto_management(self, hashes: InfoHashesOrAll, enable: bool) -> None:
        data = ParamDict.with_hashes_or_all(hashes)
        data.optional_bool("enable", enable)

        await self._request_text(
            "POST",
            "torrents/setAutoManagement",
            data=data,
        )

    async def toggle_sequential_download(self, hashes: InfoHashesOrAll) -> None:
        """Flip ``seq_dl`` values for torrents."""
        data = ParamDict.with_hashes_or_all(hashes)

        await self._request_text(
            "POST",
            "torrents/toggleSequentialDownload",
            data=data,
        )

    @virtual
    async def set_sequential_download(self, hashes: InfoHashesOrAll, value: bool) -> None:
        """
        Change ``seq_dl`` for torrents.

        .. note::

            This method is implemented by querying torrent ``seq_dl`` values, and
            :meth:`toggling <.TorrentsAPI.toggle_sequential_download>` them if needed.

        """
        torrents = await self.info(hashes=hashes)
        targets = []

        for info in torrents:
            if info.seq_dl != value:
                targets.append(info.hash)

        if targets:
            await self.toggle_sequential_download(targets)

    async def toggle_first_last_piece_prio(self, hashes: InfoHashesOrAll) -> None:
        """Flip ``f_l_piece_prio`` values for torrents."""
        await self._request_text(
            "POST",
            "torrents/toggleFirstLastPiecePrio",
            data=ParamDict.with_hashes_or_all(hashes),
        )

    @virtual
    async def set_first_last_piece_prio(self, hashes: InfoHashesOrAll, value: bool) -> None:
        """
        Change ``f_l_piece_prio`` for torrents.

        .. note::

            This method is implemented by querying torrent ``f_l_piece_prio`` values, and
            :meth:`toggling <.TorrentsAPI.toggle_first_last_piece_prio>` them if needed.

        """
        torrents = await self.info(hashes=hashes)
        targets = []

        for info in torrents:
            if info.f_l_piece_prio != value:
                targets.append(info.hash)

        if targets:
            await self.toggle_first_last_piece_prio(targets)

    async def set_force_start(self, hashes: InfoHashesOrAll, force: bool) -> None:
        """Set ``force_start`` flags for torrents."""
        data = ParamDict.with_hashes_or_all(hashes)
        data.required_bool("value", force)

        await self._request_text(
            "POST",
            "torrents/setForceStart",
            data=data,
        )

    async def set_super_seeding(self, hashes: InfoHashesOrAll, value: bool) -> None:
        """Set ``super_seeding`` flags for torrents."""
        data = ParamDict.with_hashes_or_all(hashes)
        # default value (invalid value) is treated as false
        data.required_bool("value", value)

        await self._request_text(
            "POST",
            "torrents/setSuperSeeding",
            data=data,
        )

    @overload
    async def rename_file(self, hash: InfoHash, id: int, name: str) -> None:
        """available until client 4.3.3"""

    @overload
    async def rename_file(self, hash: InfoHash, old_path: str, new_path: str) -> None:
        """available since client 4.3.3 or API 2.7.0"""

    async def rename_file(self, hash: InfoHash, *args: Any, **kwargs: Any) -> None:
        """
        Rename a file in torrent.

        On qBittorrent v4.3.3 or later, the signature is ``rename_file(hash, old_path, new_path)``.

        Below qBittorrent v4.3.3, use ``rename_file(hash, id, name)``, where ``id`` is
        the file index from :meth:`~.TorrentsAPI.files`.

        Available since qBittorrent v4.2.1 (API 2.4.0).

        Signature changed in v4.3.3 (API 2.7.0).

        See also: https://github.com/qbittorrent/qBittorrent/pull/13995

        """
        # API 2.4.0

        client = self._client()
        legacy = APIVersion.compare(client.api_version, (2, 7, 0)) < 0
        nargs = len(args)

        if nargs > 2:
            raise TypeError("Too many arguments")
        if nargs == 2:
            pass
        elif nargs == 1:
            # Missing one argument
            key = "name" if isinstance(args[0], int) else "new_path"
            try:
                args += (kwargs.pop(key),)
            except KeyError:
                raise TypeError(f"Missing argument {key!r}") from None
        else:
            try:
                if "old_path" in kwargs:
                    args = (kwargs.pop("old_path"), kwargs.pop("new_path"))
                elif "id" in kwargs:
                    args = (kwargs.pop("id"), kwargs.pop("name"))
                else:
                    raise KeyError("id" if legacy else "old_path")
            except KeyError as ex:
                raise TypeError(f"Missing argument {ex.args[0]!r}") from None

        if kwargs:
            raise TypeError(f"Extra keyword arguments: {kwargs.keys()!r}")

        arg, arg2 = args

        data = ParamDict.with_hash(hash)
        if isinstance(arg, str):
            data.required_str("oldPath", arg)
            data.required_str("newPath", arg2)
        elif isinstance(arg, int):
            data.required_int("id", arg)
            data.required_str("name", arg2)
        else:
            raise TypeError(f"Bad call signature: ({type(hash)!r}, {type(arg)!r}, {type(arg2)!r})")

        try:
            await self._request_text(
                "POST",
                "torrents/renameFile",
                data=data,
            )
        except exc.BadRequestError as ex:
            if APIVersion.compare(self._client().api_version, (2, 4, 0)) >= 0:
                note = (
                    "From qBittorrent 4.2.1, rename_file(hash, id, name) was changed"
                    " to rename_file(hash, old_path, new_path)."
                    " BadRequestError may raise if rename_file() was called"
                    " with inappropriate arguments."
                )
                exc._add_note(ex, note, logger=self._client()._logger)

            raise

    @since((2, 8, 0))
    async def rename_folder(self, hash: InfoHash, old_path: str, new_path: str) -> None:
        """Rename a folder."""
        # API 2.8.0

        data = ParamDict.with_hash(hash)
        data.required_path("oldPath", old_path)
        data.required_path("newPath", new_path)

        await self._request_text(
            "POST",
            "torrents/renameFolder",
            data=data,
        )

    async def export(self, hash: InfoHash) -> bytes:
        """Export a torrent as ``bytes``."""
        # since API v2.8.11

        params = ParamDict.with_hash(hash)
        resp = await self._client().request(
            "GET",
            "torrents/export",
            params=params,
        )

        async with resp:
            return await resp.read()


@dataclasses.dataclass
class AddFormBuilder:
    """
    Build :class:`aiohttp.FormData` used in :meth:`.TorrentsAPI.add`.

    AddFormBuilder is designed in fluent interface.
    Most of its methods return a modified copy of the builder.

    Here is an example to illustrate::

        await client.torrent.add(
            # Create a builder with a particular client
            AddFormBuilder.with_client(client)

            # Set torrent category to "linux"
            .category("linux")

            # Set ratio limit to 10
            .ratio_limit(10)

            # Add a torrent by its info hash (debian-11.7.0-amd64-netinst.iso)
            .include_url("6f84758b0ddd8dc05840bf932a77935d8b5b8b93")

            # Add a torrent by URL/magnet link (debian-11.6.0-amd64-netinst.iso)
            .include_url("magnet:?xt=urn:btih:6d4795dee70aeb88e03e5336ca7c9fcf0a1e206d")

            # Upload a torrent with its bytes data and name
            .include_url(file_bytes, "debian-12.0.0-amd64-netinst.iso")

            # Generate FormData object
            .build()
        )

    See also :APIWiki:`torrents/add <#add-new-torrent>`.

    """

    client_version: Optional[ClientVersion] = None
    api_version: Optional[APIVersion] = None

    _urls: List[str] = dataclasses.field(default_factory=list)
    _files: List[Tuple[bytes, str]] = dataclasses.field(default_factory=list)

    _savepath: Optional[str] = None
    _download_path: Optional[str] = None
    _use_download_path: Optional[bool] = None
    _cookie: Optional[str] = None
    _category: Optional[str] = None
    _tags: Optional[str] = None
    _skip_checking: Optional[bool] = None
    _paused: Optional[bool] = None
    _root_folder: Optional[bool] = None
    _rename: Optional[str] = None
    _up_limit: Optional[int] = None
    _dl_limit: Optional[int] = None
    _ratio_limit: Optional[float] = None
    _seeding_time_limit: Optional[int] = None
    _inactive_seeding_time_limit: Optional[int] = None
    _auto_tmm: Optional[bool] = None
    _sequential_download: Optional[bool] = None
    _first_last_piece_prio: Optional[bool] = None
    _add_to_top_of_queue: Optional[bool] = None
    _stop_condition: Optional[str] = None
    _content_layout: Optional[str] = None

    def __deepcopy__(self, memodict: Optional[Dict[int, Any]] = None) -> Self:
        return dataclasses.replace(
            self,
            _urls=list(self._urls),
            _files=list(self._files),
        )

    __copy__ = __deepcopy__

    @copy_self
    def include_url(self, url: str) -> Self:
        """
        Add a URL, magnet link, or info hash (SHA1/SHA256) to form.
        """

        self._urls.append(url)
        return self

    @copy_self
    def include_file(self, data: bytes, filename: Optional[str] = None) -> Self:
        """
        Add a torrent file to form.
        """

        if filename is None:
            filename = f"{len(self._files) + 1:d}.torrent"

        self._files.append((bytes(data), filename))
        return self

    def add_url(self, url: str) -> Self:
        """deprecated, use :meth:`.include_url` instead."""
        # deprecated, use include_url() instead
        return self.include_url(url)  # pragma: no cover

    def add_torrent(self, filename: str, data: bytes) -> Self:
        """deprecated, use :meth:`.include_file` instead."""
        return self.include_file(data, filename)  # pragma: no cover

    @copy_self
    def savepath(self, savepath: Optional[StrPath]) -> Self:
        """Set ``savepath`` value."""
        self._savepath = None if savepath is None else _convert_path(savepath)
        return self

    @copy_self
    def download_path(self, download_path: Optional[StrPath]) -> Self:
        """
        Set ``downloadPath`` value.

        Also use :meth:`use_download_path(True) <.use_download_path>` to enable download path.
        """
        # API v2.8.4
        self._download_path = None if download_path is None else _convert_path(download_path)
        return self

    @copy_self
    def use_download_path(self, use_download_path: Optional[bool]) -> Self:
        """
        Set ``useDownloadPath`` value.
        """
        # API v2.8.4
        self._use_download_path = use_download_path
        return self

    @copy_self
    def cookie(self, cookie: Optional[str]) -> Self:
        """Set ``cookie`` value."""
        self._cookie = cookie
        return self

    @copy_self
    def category(self, category: Optional[str]) -> Self:
        """Set ``category`` value."""
        self._category = category
        return self

    @copy_self
    def tags(self, tags: Optional[Iterable[str]]) -> Self:
        """
        Associate torrents being added with tags.

        Available since API v2.6.2.

        :param tags: list of tags.
        """
        if tags is None:
            self._tags = None
            return self

        # API 2.6.2
        _check_iterable_except_str("tags", tags)

        tags = list(tags)
        parts = []
        for item in tags:
            if not isinstance(item, str):
                raise ValueError("each tag must be a str")
            if "," in item:
                raise ValueError(f"Tag cannot contain comma: {item!r}")
            parts.append(item)

        self._tags = ",".join(parts)
        return self

    @copy_self
    def skip_checking(self, skip_checking: Optional[bool]) -> Self:
        """Set ``skip_checking`` value."""
        self._skip_checking = skip_checking
        return self

    @copy_self
    def paused(self, paused: Optional[bool]) -> Self:
        """Set ``paused`` value."""
        self._paused = paused
        return self

    @copy_self
    def root_folder(self, root_folder: Optional[bool]) -> Self:
        """
        Set ``root_folder`` value.

        Removed on qBittorrent v4.3.2 and later. Use :meth:`.content_layout` instead.
        """
        self._root_folder = root_folder
        return self

    @copy_self
    def rename(self, rename: Optional[str]) -> Self:
        """Set ``rename`` value, which is the new torrent name."""
        self._rename = rename
        return self

    @copy_self
    def up_limit(self, up_limit: Optional[int]) -> Self:
        """Set ``upLimit`` in bytes/s"""
        self._up_limit = up_limit
        return self

    @copy_self
    def dl_limit(self, dl_limit: Optional[int]) -> Self:
        """Set ``dlLimit`` in bytes/s"""
        self._dl_limit = dl_limit
        return self

    @copy_self
    @since((2, 8, 1))
    def ratio_limit(self, ratio_limit: Union[float, RatioLimits, None]) -> Self:
        """Set ``ratioLimit`` value."""
        # API 2.8.1
        self._ratio_limit = None if ratio_limit is None else float(ratio_limit)
        return self

    @copy_self
    @since((2, 8, 1))
    def seeding_time_limit(
        self,
        seeding_time_limit: Union[timedelta, Minutes, SeedingTimeLimits, None],
    ) -> Self:
        """Set ``seedingTimeLimit`` value."""
        # API 2.8.1
        if seeding_time_limit is None:
            self._seeding_time_limit = None
        else:
            self._seeding_time_limit = int(_convert_duration(seeding_time_limit, TimeUnit.MINUTES))
        return self

    @copy_self
    def inactive_seeding_time_limit(
        self,
        inactive_seeding_time_limit: Union[timedelta, Minutes, SeedingTimeLimits, None],
    ) -> Self:
        """Set ``inactiveSeedingTimeLimit`` value."""
        # API 2.9.2
        if inactive_seeding_time_limit is None:
            self._inactive_seeding_time_limit = None
        else:
            self._inactive_seeding_time_limit = int(
                _convert_duration(inactive_seeding_time_limit, TimeUnit.MINUTES)
            )
        return self

    @copy_self
    def auto_tmm(self, auto_tmm: Optional[bool]) -> Self:
        """Set ``autoTMM`` value."""
        self._auto_tmm = auto_tmm
        return self

    @copy_self
    def sequential_download(self, sequential_download: Optional[bool]) -> Self:
        """Set ``sequentialDownload`` value."""
        self._sequential_download = sequential_download
        return self

    @copy_self
    def first_last_piece_prio(self, first_last_piece_prio: Optional[bool]) -> Self:
        """Set ``firstLastPiecePrio`` value."""
        self._first_last_piece_prio = first_last_piece_prio
        return self

    @copy_self
    def add_to_top_of_queue(self, add_to_top_of_queue: Optional[bool]) -> Self:
        """Set ``addToTopOfQueue`` value"""
        # found in v4.6.0, API 2.8.19
        self._add_to_top_of_queue = add_to_top_of_queue
        return self

    @copy_self
    def stop_condition(self, stop_condition: Optional[StopCondition]) -> Self:
        """Set ``stopCondition`` value."""
        # API v2.8.15
        self._stop_condition = None if stop_condition is None else str(stop_condition)
        return self

    @copy_self
    def content_layout(self, content_layout: Optional[ContentLayout]) -> Self:
        """Set ``contentLayout`` value."""
        # API v2.7.0
        self._content_layout = None if content_layout is None else str(content_layout)
        return self

    def build(self) -> aiohttp.FormData:
        """
        Build :class:`~aiohttp.FormData`.
        """

        def bool_str(b: bool) -> str:
            """boolean to lowercase string"""
            return "true" if b else "false"

        form = aiohttp.FormData()

        if self._urls:
            content_type = None
            if APIVersion.compare(self.api_version, (2, 6, 0)) < 0:
                # Force multipart/form-data with content_type
                # Workaround issue in qBittorrent 4.1 (4.2?)
                # See https://github.com/qbittorrent/qBittorrent/pull/10458
                content_type = "text/plain"

            form.add_field("urls", "\n".join(self._urls), content_type=content_type)

        for filedata, filename in self._files:
            form.add_field(
                "torrents",
                filedata,
                filename=filename,
                content_type="application/x-bittorrent",
            )

        if self._savepath is not None:
            form.add_field("savepath", self._savepath)

        if self._download_path is not None:
            form.add_field("downloadPath", self._download_path)

        if self._use_download_path is not None:
            form.add_field("useDownloadPath", bool_str(self._use_download_path))

        if self._cookie is not None:
            form.add_field("cookie", self._cookie)

        if self._category is not None:
            form.add_field("category", self._category)

        if self._tags is not None:
            form.add_field("tags", self._tags)

        if self._skip_checking is not None:
            form.add_field("skip_checking", bool_str(self._skip_checking))

        if self._paused is not None:
            form.add_field("paused", bool_str(self._paused))

        if self._root_folder is not None:
            form.add_field("root_folder", bool_str(self._root_folder))

        if self._rename is not None:
            form.add_field("rename", self._rename)

        if self._up_limit is not None:
            form.add_field("upLimit", str(self._up_limit))

        if self._dl_limit is not None:
            form.add_field("dlLimit", str(self._dl_limit))

        if self._ratio_limit is not None:
            form.add_field("ratioLimit", str(self._ratio_limit))

        if self._seeding_time_limit is not None:
            form.add_field("seedingTimeLimit", str(self._seeding_time_limit))

        if self._inactive_seeding_time_limit is not None:
            form.add_field("inactiveSeedingTimeLimit", str(self._inactive_seeding_time_limit))

        if self._auto_tmm is not None:
            form.add_field("autoTMM", bool_str(self._auto_tmm))

        if self._sequential_download is not None:
            form.add_field("sequentialDownload", bool_str(self._sequential_download))

        if self._add_to_top_of_queue is not None:
            form.add_field("addToTopOfQueue", bool_str(self._add_to_top_of_queue))

        if self._first_last_piece_prio is not None:
            form.add_field("firstLastPiecePrio", bool_str(self._first_last_piece_prio))

        if self._stop_condition is not None:
            form.add_field("stopCondition", self._stop_condition)

        if self._content_layout is not None:
            form.add_field("contentLayout", self._content_layout)

        return form

    @classmethod
    def with_client(cls, client: "APIClient") -> Self:
        """
        Return :class:`.AddFormBuilder` to build :class:`~aiohttp.FormData`
        used in :meth:`.TorrentsAPI.add`.

        :param client: :class:`.APIClient`.
        :return: :class:`.AddFormBuilder`.
        """

        return cls(
            client_version=client.client_version,
            api_version=client.api_version,
        )


def _convert_duration(
    delta: Union[timedelta, float],
    unit: TimeUnit,
) -> float:
    """
    Convert duration to float

    timedelta is converted to the unit or seconds if not specified.
    int/float is returned unchanged.
    """
    if isinstance(delta, timedelta):
        delta = unit.from_seconds(delta.total_seconds())

        if TYPE_CHECKING:
            assert isinstance(delta, (int, float))

    return delta


def _convert_path(path: StrPath) -> str:
    """Convert path-like objects to POSIX path str"""
    return os.fsdecode(path).replace("\\", "/")



import asyncio

from aioqbt.api import AddFormBuilder
from aioqbt.client import create_client


async def main():
    client = await create_client(
        "http://localhost:8080/api/v2/",
        username="admin",
        password="adminadmin",
    )

    async with client:
        print("Version", await client.app.version())
        print("API", await client.app.webapi_version())

        # Add ubuntu-22.04.3-desktop-amd64.iso
        url = "https://releases.ubuntu.com/22.04/ubuntu-22.04.3-desktop-amd64.iso.torrent"
        await client.torrents.add(AddFormBuilder.with_client(client).include_url(url).build())

        await asyncio.sleep(10)  # wait a few seconds

        torrents = await client.torrents.info()
        for info in torrents:
            print(info)


if __name__ == "__main__":
    asyncio.run(main())



import logging
import sys

import pytest

from aioqbt import exc


def test_add_note(caplog: pytest.LogCaptureFixture):
    name = __name__
    logger = logging.getLogger(name)
    note = "**note_message**"

    try:
        try:
            raise ValueError
        except ValueError as ex:
            exc._add_note(ex, note, logger=logger)
            raise
    except ValueError as value_error:
        if sys.version_info >= (3, 11):
            assert any(note in s for s in value_error.__notes__), value_error.__notes__
        else:
            assert any(note in s for s in caplog.messages), caplog.messages



import re
from functools import total_ordering
from typing import Any, NamedTuple, Optional, Tuple, Union

from typing_extensions import Protocol, Self

from aioqbt.exc import VersionError

__all__ = (
    "ClientVersion",
    "APIVersion",
)

_VERSION_PATTERN = re.compile(
    r"^v?(\d+)\.(\d+)(?:\.(\d+)(?:\.(\d+))?)?((?:alpha|beta|rc)\d*)?$",
    re.IGNORECASE,
)

_API_VERSION_PATTERN = re.compile(
    r"^(\d+)\.(\d+)(?:\.(\d+))?$",
    re.IGNORECASE,
)


@total_ordering
class ClientVersion:
    """
    Represent client version.
    """

    _key: Tuple[int, int, int, int, str, int]
    _status: str

    def __init__(self, major: int, minor: int, patch: int, build: int = 0, status: str = ""):
        st0, st1 = self._parse_status(status)
        self._key = (major, minor, patch, build, st0, st1)
        self._status = status

    @property
    def major(self) -> int:
        """Major number."""
        return self._key[0]

    @property
    def minor(self) -> int:
        """Minor number."""
        return self._key[1]

    @property
    def patch(self) -> int:
        """Patch number."""
        return self._key[2]

    @property
    def build(self) -> int:
        """Build number."""
        return self._key[3]

    @property
    def status(self) -> str:
        """Status string."""
        return self._status

    def __eq__(self, other: object) -> bool:
        if isinstance(other, ClientVersion):
            return self._key == other._key
        return NotImplemented

    def __hash__(self) -> int:
        return hash(self._key)

    def __lt__(self, other: Any) -> bool:
        if isinstance(other, ClientVersion):
            return self._key < other._key
        return NotImplemented

    def __str__(self) -> str:
        result = f"{self.major:d}.{self.minor:d}.{self.patch:d}"

        if self.build != 0:
            result += f".{self.build:d}"

        result += self.status
        return result

    @classmethod
    def _parse_status(cls, status: str) -> Tuple[str, int]:
        if status == "":
            return "release", 0

        match = re.match(r"(alpha|beta|rc)(\d*)", status, re.IGNORECASE)
        if match is None:
            raise ValueError(f"Bad status: {status!r}")
        a, b = match.groups()
        return a.lower(), int(b or 0)

    @classmethod
    def parse(cls, version: str) -> "ClientVersion":
        """
        Parse client version.

        Format::

            major.minor.patch[.build][status]

        Examples::

            4.2.5
            4.4.0beta2
            4.4.3.1
        """

        match = _VERSION_PATTERN.match(version)
        if match is None:
            raise ValueError(f"Bad version: {version!r}")

        major = int(match[1])
        minor = int(match[2])
        patch = int(match[3] or 0)
        build = int(match[4] or 0)
        status = match[5] or ""
        return cls(major, minor, patch, build, status)


class APIVersion(NamedTuple):
    """
    Represent API version.

    Instances can also be compared with 3-tuple of :class:`int`.

    """

    major: int
    """Major number."""

    minor: int
    """Minor number."""

    release: int = 0
    """Release number."""

    @classmethod
    def parse(cls, version: str) -> "APIVersion":
        """
        Parse API version.

        Format::

            major.minor[.release]

        where ``major``, ``minor`` and ``release`` are all digits.
        """
        match = _API_VERSION_PATTERN.match(version)

        if match is None:
            raise ValueError(f"Bad API version: {version!r}")

        s1, s2, s3 = match.groups()

        return cls(int(s1), int(s2), int(s3 or 0))

    def __str__(self) -> str:
        return f"{self.major}.{self.minor}.{self.release}"

    @classmethod
    def compare(
        cls,
        a: Optional[Union[Self, Tuple[int, int, int]]],
        b: Optional[Union[Self, Tuple[int, int, int]]],
    ) -> int:
        """
        Compare two API versions.

        Return zero if ``a == b``; a negative value if ``a < b``;
        or a positive value if ``a > b``.

        ``None`` is a special value treated as the latest version.

        :return: integer value indicating version relationship.
        """

        if a is None:
            if b is None:
                return 0
            else:
                return 1
        elif b is None:
            return -1
        elif a == b:
            return 0
        elif a < b:
            return -1
        else:
            return 1


class Comparable(Protocol):
    def __eq__(self, other: Any) -> bool:
        pass

    def __ne__(self, other: Any) -> bool:
        pass

    def __lt__(self, other: Any) -> bool:
        pass

    def __le__(self, other: Any) -> bool:
        pass

    def __gt__(self, other: Any) -> bool:
        pass

    def __ge__(self, other: Any) -> bool:
        pass


def version_satisfy(version: Optional[Comparable], minimum: Comparable) -> bool:
    """
    Compare version with minimum requirement and return boolean

    :param version: current version, or ``None`` as the latest
    :param minimum: minimum version
    """
    return version is None or version >= minimum


def version_check(version: Optional[Comparable], minimum: Comparable) -> None:
    """
    Compare version with minimum requirement and raise if violated.

    :param version: current version, or ``None`` as the latest
    :param minimum: minimum version
    """
    if version is not None and version < minimum:
        raise VersionError(f"Version {minimum} is required but {version} is found")


def param_version_check(param: str, version: Optional[Comparable], minimum: Comparable) -> None:
    if not version_satisfy(version, minimum):
        if type(minimum) is tuple:
            minimum = type(version)(*minimum)  # type: ignore

        raise VersionError(f"{param!r} requires version {version!r} but {minimum!r} found")



import random
from hashlib import sha1
from typing import Any, List, Mapping, NamedTuple, Optional, Tuple

from helper import bencode
from helper.lang import randbytes


class TorrentData(NamedTuple):
    hash: str
    name: str
    data: bytes

    @property
    def magnet(self) -> str:
        return f"magnet:?xt=urn:btih:{self.hash}"


class PieceHasher:
    def __init__(self, piece_length: int):
        if piece_length <= 0:
            raise ValueError(f"piece_length = {piece_length!r}")

        self._finalized = False
        self._pieces: List[bytes] = []
        self._piece_length = piece_length
        self._hasher = sha1()
        self._hasher_rem = piece_length

    def _consume(self, data: bytes, final: bool) -> None:
        if self._finalized:
            raise ValueError("finalize() has been called")

        pos = 0
        remaining = len(data)

        while remaining > 0:
            consumed = min(remaining, self._hasher_rem)
            self._hasher.update(data[pos : pos + consumed])
            pos += consumed
            remaining -= consumed
            self._hasher_rem -= consumed

            if self._hasher_rem == 0:
                self._pieces.append(self._hasher.digest())
                self._hasher = sha1()
                self._hasher_rem = self._piece_length

        if not final:
            return

        self._finalized = True
        if self._hasher_rem < self._piece_length:
            self._pieces.append(self._hasher.digest())

    def update(self, data: bytes) -> None:
        self._consume(data, False)

    def finalize(self, data: bytes = b"") -> List[bytes]:
        self._consume(data, True)
        return self._pieces

    @property
    def pieces(self) -> List[bytes]:
        return self._pieces

    @property
    def piece_length(self) -> int:
        return self._piece_length


def _rand_and_message(name: str) -> Tuple[random.Random, str]:
    r = random.Random(name)
    message = randbytes(r, 16).hex()
    return r, message


def make_torrent_single(
    name: str,
    torrent: Optional[Mapping[str, Any]] = None,
) -> TorrentData:
    piece_length = 16 * 1024
    torrent = dict(torrent) if torrent else {}

    r, message = _rand_and_message(name)
    data = message.encode("latin-1")
    pieces = PieceHasher(piece_length).finalize(data)

    info_dict = {
        "length": len(data),
        "name": name,
        "piece length": piece_length,
        "pieces": b"".join(pieces),
    }
    info_hash = sha1(bencode.dumps(info_dict)).hexdigest()
    torrent["info"] = info_dict
    torrent_bytes = bencode.dumps(torrent)

    return TorrentData(
        hash=info_hash,
        name=name,
        data=torrent_bytes,
    )


def make_torrent_files(
    name: str,
    torrent: Optional[Mapping[str, Any]] = None,
) -> TorrentData:
    piece_length = 16 * 1024
    torrent = dict(torrent) if torrent else {}

    r, message = _rand_and_message(name)

    # list of (file path, file content)
    contents: List[Tuple[str, bytes]] = []

    for i in range(5):
        file_path = f"files/{i:02d}.txt"
        file_data = randbytes(r, piece_length // 2).hex().encode("latin-1")
        contents.append((file_path, file_data))

    files = []
    piece_hasher = PieceHasher(piece_length)

    for file_path, file_data in contents:
        files.append(
            {
                "path": file_path.split("/"),
                "length": len(file_data),
            }
        )
        piece_hasher.update(file_data)

    pieces = piece_hasher.finalize()

    info_dict = {
        "name": name,
        "files": files,
        "piece length": piece_length,
        "pieces": b"".join(pieces),
    }
    info_hash = sha1(bencode.dumps(info_dict)).hexdigest()
    torrent["info"] = info_dict
    torrent_bytes = bencode.dumps(torrent)

    return TorrentData(
        hash=info_hash,
        name=name,
        data=torrent_bytes,
    )



import os
import re
from datetime import timedelta
from math import isfinite
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    Iterator,
    Mapping,
    MutableMapping,
    NoReturn,
    Optional,
    TypeVar,
    Union,
)

from typing_extensions import Self

from aioqbt.bittorrent import InfoHash, InfoHashes, InfoHashesOrAll, _info_hash_str
from aioqbt.chrono import TimeUnit
from aioqbt.typing import StrPath

T = TypeVar("T")
PrepareFn = Callable[[T], Union[float, str]]

_CAMEL_PATTERN = re.compile(r"(?!^)([A-Z]+)", re.ASCII)


def _camel2snake(name: str) -> str:
    """
    Convert camelCase to snake_case

    Examples:
    - helloWorld -> hello_world
    - SendHTTP -> send_http
    - HTTPRequest -> httprequest
    - HTTP2Request -> http2_request
    - seedingTimeLimit -> seeding_time_limit
    """
    return _CAMEL_PATTERN.sub(lambda m: f"_{m.group(1).lower()}", name).lower()


def _param_name(key: str, param: Optional[str]) -> str:
    """
    Derive parameter name from key name
    """

    if param is None:
        return _camel2snake(key)
    else:
        return param


def _missing() -> NoReturn:
    raise AssertionError


class ParamDict(MutableMapping[str, str]):
    """
    A helper dict to construct GET params and POST data in common pattern
    """

    _data: Dict[str, str]

    def __init__(
        self,
        data: Optional[Mapping[str, Any]] = None,
    ):
        if data is None:
            self._data = {}
        elif isinstance(data, ParamDict):
            self._data = data._data.copy()
        else:
            self._data = {}

            for k, v in list(data.items()):
                self.put(k, v)

    def __setitem__(self, key: str, value: str) -> None:
        self._data[key] = value

    def __delitem__(self, key: str) -> None:
        del self._data[key]

    def __getitem__(self, key: str) -> str:
        return self._data[key]

    def __len__(self) -> int:
        return len(self._data)

    def __iter__(self) -> Iterator[str]:
        return iter(self._data)

    def __deepcopy__(self, memo: Optional[Dict[int, Any]] = None) -> Self:
        return type(self)(self)

    __copy__ = __deepcopy__

    @classmethod
    def _missing_param(cls, key: str, param: Optional[str]) -> TypeError:
        return TypeError(f"{_param_name(key, param)!r} is required")

    def to_dict(self) -> Dict[str, str]:
        return self._data.copy()

    def _put(
        self,
        key: str,
        param: Optional[str],
        value: Any,
        optional: bool,
        prepare: Optional[PrepareFn[Any]] = None,
        default: Any = _missing,
    ) -> None:
        """
        Associate a key with a value

        The put operation is ignored if the value is None and optional is True.
        An exception is raised if the value is None but optional is False.

        :param key: the key name.
        :param value: the value or None if empty.
        :param param: the parameter name.
        :param optional: whether value None is allowed.
        :param prepare: callable to convert non-None value to str, int, or float
        :param default: default value if value is None.
        """
        if value is None:
            if default is _missing:
                if optional:
                    return

                raise self._missing_param(key, param)

            value = default

        if prepare is None:
            if not isinstance(value, (str, int, float)):
                raise TypeError(
                    f"{_param_name(key, param)!r} expect"
                    f" str, int, or float instead of {type(value)}"
                )

            value = str(value)
        else:
            value = prepare(value)

            if not isinstance(value, (str, int, float)):
                raise TypeError(
                    f"{_param_name(key, param)!r} expect {prepare} result in"
                    f" str, int, or float instead of {type(value)}"
                )

            value = str(value)

        self._data[key] = value

    def put(
        self,
        key: str,
        value: Any,
        *,
        param: Optional[str] = None,
        optional: bool = False,
        prepare: Optional[PrepareFn[Any]] = None,
        default: Any = _missing,
    ) -> None:
        self._put(key, param, value, optional, prepare, default)

    def optional_str(self, key: str, value: Optional[str], *, param: Optional[str] = None) -> None:
        self._put(key, param, value, True, str)

    def required_str(self, key: str, value: str, *, param: Optional[str] = None) -> None:
        self._put(key, param, value, False, str)

    def optional_int(self, key: str, value: Optional[int], *, param: Optional[str] = None) -> None:
        self._put(key, param, value, True, int)

    def required_int(self, key: str, value: int, *, param: Optional[str] = None) -> None:
        self._put(key, param, value, False, int)

    def optional_float(
        self, key: str, value: Optional[float], *, param: Optional[str] = None
    ) -> None:
        self._put(key, param, value, True, float)

    def required_float(self, key: str, value: float, *, param: Optional[str] = None) -> None:
        self._put(key, param, value, False, float)

    def optional_bool(
        self, key: str, value: Optional[bool], *, param: Optional[str] = None
    ) -> None:
        self._put(key, param, value, True, _prepare_bool)

    def required_bool(self, key: str, value: bool, *, param: Optional[str] = None) -> None:
        self._put(key, param, value, False, _prepare_bool)

    def _put_duration(
        self,
        key: str,
        param: Optional[str],
        value: Union[timedelta, int, float, None],
        unit: TimeUnit,
        optional: bool,
    ) -> None:
        if isinstance(value, timedelta):
            value = unit.from_seconds(value.total_seconds())

        if isinstance(value, float) and not isfinite(value):
            raise ValueError(f"{_param_name(key, param)!r} expect a finite value: {value!r}")

        self._put(key, param, value, optional, int)

    def required_duration(
        self,
        key: str,
        value: Union[timedelta, int, float],
        unit: TimeUnit,
        *,
        param: Optional[str] = None,
    ) -> None:
        self._put_duration(key, param, value, unit, False)

    def optional_duration(
        self,
        key: str,
        value: Union[timedelta, int, float, None],
        unit: TimeUnit,
        *,
        param: Optional[str] = None,
    ) -> None:
        self._put_duration(key, param, value, unit, True)

    def required_path(
        self,
        key: str,
        value: StrPath,
        *,
        param: Optional[str] = None,
    ) -> None:
        self._put(key, param, value, False, _prepare_path)

    def optional_path(
        self,
        key: str,
        value: Optional[StrPath],
        *,
        param: Optional[str] = None,
    ) -> None:
        self._put(key, param, value, True, _prepare_path)

    def _put_list(
        self,
        key: str,
        value: Optional[Iterable[T]],
        sep: str,
        param: Optional[str],
        optional: bool,
        prepare: Optional[PrepareFn[T]],
        nonempty: bool,
    ) -> None:
        if value is None:
            if optional:
                return

            raise self._missing_param(key, param)

        if prepare is None:
            items = [str(s) for s in value]
        else:
            items = [str(prepare(s)) for s in value]

        if nonempty and not items:
            raise ValueError(f"{_param_name(key, param)!r} must not be empty")

        self._data[key] = sep.join(items)

    def required_list(
        self,
        key: str,
        value: Iterable[T],
        sep: str,
        *,
        param: Optional[str] = None,
        prepare: Optional[PrepareFn[T]] = None,
        nonempty: bool = False,
    ) -> None:
        self._put_list(key, value, sep, param, False, prepare, nonempty)

    def optional_list(
        self,
        key: str,
        value: Optional[Iterable[T]],
        sep: str,
        *,
        param: Optional[str] = None,
        prepare: Optional[PrepareFn[T]] = None,
        nonempty: bool = False,
    ) -> None:
        self._put_list(key, value, sep, param, True, prepare, nonempty)

    @classmethod
    def with_hash(
        cls,
        hash: InfoHash,
        *,
        key: Optional[str] = None,
        param: Optional[str] = None,
    ) -> Self:
        if key is None:
            key = "hash"

        res = cls()
        res.put(key, hash, param=param, prepare=_info_hash_str)
        return res

    @classmethod
    def with_hashes(
        cls,
        hashes: InfoHashes,
        *,
        key: Optional[str] = None,
        param: Optional[str] = None,
        nonempty: bool = False,
    ) -> Self:
        if key is None:
            key = "hashes"

        res = cls()
        res.required_list(key, hashes, "|", param=param, prepare=_info_hash_str, nonempty=nonempty)
        return res

    @classmethod
    def with_hashes_or_all(
        cls,
        hashes: InfoHashesOrAll,
        *,
        key: Optional[str] = None,
        param: Optional[str] = None,
        nonempty: bool = False,
    ) -> Self:
        if key is None:
            key = "hashes"

        res = cls()
        if hashes == "all":
            res.put(key, "all", param=param)
        else:
            res.required_list(
                key,
                hashes,
                "|",
                param=param,
                prepare=_info_hash_str,
                nonempty=nonempty,
            )
        return res


def _prepare_bool(b: bool) -> str:
    """
    Convert boolean values to "true" or "false" string in lowercase
    """
    return "true" if bool(b) else "false"


def _prepare_path(p: StrPath) -> str:
    """
    Convert path-like object to str
    """
    return os.fsdecode(p).replace("\\", "/")



import pytest

from aioqbt.exc import VersionError
from aioqbt.version import (
    APIVersion,
    ClientVersion,
    param_version_check,
    version_check,
    version_satisfy,
)


def test_client_version():
    string = "4.2.5.1"
    version = ClientVersion.parse(string)

    assert version.major == 4
    assert version.minor == 2
    assert version.patch == 5
    assert version.build == 1
    assert version.status == ""

    assert version == version
    assert str(version) == string
    assert version != object()
    assert {version: version}[version] == version

    with pytest.raises(TypeError):
        assert version < object()

    string2 = "4.2.5rc1"
    version2 = ClientVersion.parse(string2)
    assert version2.status == "rc1"
    assert str(version2) == "4.2.5rc1"
    assert version2 < version

    with pytest.raises(ValueError):
        ClientVersion.parse("4.2.x")

    with pytest.raises(ValueError):
        ClientVersion.parse("4.2.5bad")

    with pytest.raises(ValueError):
        ClientVersion(4, 2, 5, 0, "bad")


def test_api_version():
    version = APIVersion.parse("2.2.0")
    assert str(version) == "2.2.0"

    version2 = APIVersion.parse("2.2.5")
    assert version < version2

    with pytest.raises(ValueError):
        APIVersion.parse("2.2.5beta1")


def test_api_version_compare():
    a = APIVersion(4, 5, 1)

    assert APIVersion.compare(None, None) == 0
    assert APIVersion.compare(a, a) == 0
    assert APIVersion.compare(None, a) > 0
    assert APIVersion.compare(a, None) < 0

    b = APIVersion(4, 6, 1)
    assert APIVersion.compare(a, b) < 0
    assert APIVersion.compare(b, a) > 0


def test_version_utilities():
    v1 = APIVersion.parse("2.2.0")

    assert version_satisfy(v1, v1)
    assert version_satisfy(None, v1)

    v2 = APIVersion.parse("2.4.0")
    assert not version_satisfy(v1, v2)
    assert version_satisfy(v2, v1)

    with pytest.raises(VersionError):
        version_check(v1, v2)

    param_version_check("my_param", v1, v1)
    param_version_check("my_param", v2, v1)

    with pytest.raises(VersionError):
        param_version_check("param", v1, v2)


def test_param_version_check():
    v = APIVersion.parse("2.2.5")

    with pytest.raises(VersionError, match="parameter"):
        param_version_check("parameter", v, (2, 4, 5))



import copy
from functools import wraps
from typing import Any, Callable, Optional, TypeVar, overload

T = TypeVar("T")
Fn = TypeVar("Fn", bound=Callable[..., Any])


@overload
def copy_self(fn: None) -> Callable[[Fn], Fn]:
    pass


@overload
def copy_self(fn: Fn) -> Fn:
    pass


def copy_self(fn: Optional[Callable[..., T]] = None) -> Any:
    """
    Copy the first argument before execute the wrapped function
    """
    if fn is None:
        return copy_self

    @wraps(fn)
    def wrapper(self: T, *args: Any, **kwargs: Any) -> T:
        return fn(copy.copy(self), *args, **kwargs)

    return wrapper



from enum import Enum

from aioqbt._compat import repr_enum
from aioqbt.api.types import FilePriority, PieceState, TorrentState


def test_repr_enum():
    @repr_enum
    class StrEnum(str, Enum):
        pass

    class Strings(StrEnum):
        HELLO = "hello"

    assert str(Strings.HELLO) == "hello"

    @repr_enum
    class IntEnum(int, Enum):
        pass

    class Numbers(IntEnum):
        ONE = 1

    assert str(Numbers.ONE) == "1"


def check_enum(cls):
    for name, member in cls.__members__.items():
        assert str(member) == str(member._value_), repr(member)


def test_api_enums():
    enum_types = (
        TorrentState,
        PieceState,
        FilePriority,
    )

    for cls in enum_types:
        check_enum(cls)
